'''
General utilities 
'''


from .mprw import ModpathRW
def multipackage(initmethod):
    '''
    Wrapper for init methods of packages
    allowing multiple specifications. 

    Manages the storage of multiple instaces
    of the same package on the parent object
    '''

    def wrapper(self, *args, **kwargs):
        # Check model 
        model = args[0]
        if ( not isinstance( model, ModpathRW ) ):
            raise TypeError(
                self.__class__.__name__ + ':' + 
                ' Invalid parent model class. It should be ModpathRW, ' + 
                str(type(model)) + ' was given.'
            )
        # Create the multipackage attribute if not present 
        if ( not hasattr( model, 'multipackage' ) ):
            model.multipackage = {}
        # Get ftype and initialize multipackage dict
        ftype = self._ftype()
        try:
            model.multipackage[ftype]
        except KeyError:
            model.multipackage[ftype] = {
                    'unitnumber': model.next_unit(),
                    'count'     : 0,
                    'instances' : [],
                }

        # Give an id and call initmethod
        # Makes the id consistent with fortran , one-based
        self.id = model.multipackage[ftype]['count'] + 1 
        met     = initmethod(self, *args, **kwargs)

        # If call was successful, then add the package to the parent 
        # and increment counter. If there was an exception, the instance
        # is not stored and everything remains consistent.

        # The difference here w.r.t mf6, is that although there are
        # "multiple" package instances, for modpath-rw is still only
        # one input package definition, containing multiple specifications.
        # So it needs to add the package to the parent only the first
        # time and then store the individual instances somewhere.

        # Add package to the parent only the first time
        if self.parent.multipackage[ftype]['count'] == 0:
            self.parent.add_package(self)
        # Save instance
        self.parent.multipackage[ftype]['instances'].append( self )
        self.parent.multipackage[ftype]['count'] += 1

        return met

    return wrapper



'''
Configuration of MODPATH-RW particle groups

Update particle groups classes from modpath7.
Extend interfaces to understand mass and solute
'''

# flopy
from flopy.modpath.mp7particlegroup import (
    ParticleGroup,
    ParticleGroupLRCTemplate,
    ParticleGroupNodeTemplate,
)
mp7ParticleGroup = ParticleGroup
mp7ParticleGroupLRCTemplate = ParticleGroupLRCTemplate
mp7ParticleGroupNodeTemplate = ParticleGroupNodeTemplate


class ParticleGroup( mp7ParticleGroup ): 

    def __init__( self, *args, mass=1.0, solute=0, **kwargs ):
        # Call parent constructor
        super().__init__(*args,**kwargs) 
        
        if mass != 0:
            self.mass = mass
        else:
            raise ValueError(
                self.__class__.__name__ + ':' + 
                ' Given mass for the particle group is zero. '
            )
        self.solute = solute


    def write( self, fp=None, ws=".", mass=False, solute=False): 
        # Call base class write method to write common data
        super().write(fp, ws)

        if mass:
            # Write the particle mass
            fp.write(f"{self.mass:.16f}\n")
        if solute:
            # Write the solute id 
            fp.write(f"{self.solute:9d}\n")


        return


class ParticleGroupLRCTemplate( mp7ParticleGroupLRCTemplate ): 

    def __init__( self, *args, mass=1.0, solute=0, **kwargs ):
        # Call parent constructor
        super().__init__(*args,**kwargs) 
        
        if mass != 0:
            self.mass = mass
        else:
            raise ValueError(
                self.__class__.__name__ + ':' + 
                ' Given mass for the particle group is zero. '
            )
        self.solute = solute


    def write( self, fp=None, ws=".", mass=False, solute=False): 
        # Call base class write method to write common data
        super().write(fp, ws)

        if mass:
            # Write the particle mass
            fp.write(f"{self.mass:.16f}\n")
        if solute:
            # Write the solute id 
            fp.write(f"{self.solute:9d}\n")


        return


class ParticleGroupNodeTemplate( mp7ParticleGroupNodeTemplate ): 

    def __init__( self, *args, mass=1.0, solute=0, **kwargs ):
        # Call parent constructor
        super().__init__(*args,**kwargs) 
        
        if mass != 0:
            self.mass = mass
        else:
            raise ValueError(
                self.__class__.__name__ + ':' + 
                ' Given mass for the particle group is zero. '
            )
        self.solute = solute


    def write( self, fp=None, ws=".", mass=False, solute=False): 
        # Call base class write method to write common data
        super().write(fp, ws)

        if mass:
            # Write the particle mass
            fp.write(f"{self.mass:.16f}\n")
        if solute:
            # Write the solute id 
            fp.write(f"{self.solute:9d}\n")


        return



'''
Configuration of MODPATH-RW dispersion
'''

# python
import numpy as np
from enum import Enum

# flopy
from flopy.pakbase import Package
from flopy.utils import Util3d

# local 
from .utils import multipackage

class inputFormat(Enum): 
    '''
    Enumerate input formats for the package
    '''
    uni         = 0
    uniform     = 0
    dist        = 1
    distributed = 1
    u3d         = 1


class dispersionModel(Enum): 
    '''
    Enumerate input formats for the package
    '''
    iso          = 0
    isotropic    = 0
    axi          = 1
    axisymmetric = 1


class ModpathRWDsp( Package ):
    """
    MODPATH-RW Dispersion Package Class.

    Parameters
    ----------
    model : model object
        The model object (of type :class: ModpathRW) to which
        this package will be added.
    modelkind : str
        The dispersion model. Determines interpretation of dispersivities. 
        Allowed values are: 
          * iso/isotropic: transverse dispersion is isotropic, interpret parameters
            alphal, alphat. 
          * axi/axisymmetric: the dispersion model from Lichtner et al. 2002 for 
            vertical symmetry axis. Interpret four dispersivities: alphalh, alphalv,
            alphath and alphatv. Index h means perpendicular to the axis of symmetry 
            and index v means parallel to the axis of symmetry.
    inputformat : str
        The format for writing dispersion model paramters. Allowed values are: 
          * uni/uniform: Parameters are spatially uniform
          * dist/distributed/u3d: Parameters are spatially distributed
        Notes:
         * The inputformat distinction determines the reader to be used in MODPATH-RW.
           In the case of spatially uniform parameters, arrays are allocated with only 
           one value instead of the default (one value per cell) while using u3d readers.
         * If None is given, it will infer an inputformat from the shape of dispersion parameters.
    alphal : float, np.ndarray
        Longitudinal dispersivity. Should satisfy alphal >=0.
        Interpreted when modelkind = iso/isotropic
    alphat : float, np.ndarray
        Transverse dispersivity. Should satisfy alphat >=0.
        Interpreted when modelkind = iso/isotropic
    alphalh : float, np.ndarray
        Longitudinal dispersivity for horizontal flow. Should satisfy alphalh >=0.
        Interpreted when modelkind = axi/axisymmetric
    alphalv : float, np.ndarray
        Longitudinal dispersivity for vertical flow. Should satisfy alphalv >=0.
        Interpreted when modelkind = axi/axisymmetric
    alphath : float, np.ndarray
        Horizontal transverse dispersivity. Should satisfy alphath >=0.
        Interpreted when modelkind = axi/axisymmetric
    alphatv : float, np.ndarray
        Vertical transverse dispersivity. Should satisfy alphatv >=0.
        Interpreted when modelkind = axi/axisymmetric
    dmeff  : float, np.array
        Effective molecular diffusion, corrected by tortuosity. Should satisfy dmeff >=0.
    stringid : str, optional
        An id for the dispersion specification. By default concatenates
        package ftype and integer id. 
    extension : str, optional
        File extension (default is 'dsp').
    """


    @staticmethod
    def _ftype():
        return 'DSP'


    @multipackage
    def __init__(
        self,
        model,
        modelkind   = 'iso'   , # iso/isotropic, axi/axisymmetric
        inputformat = None    , # uni/uniform or dist/distributed/u3d
        alphal      = 0.1     , # xx
        alphat      = 0.01    , # yy, zz isotropic transverse dispersivity
        dmeff       = 0.0     , # corrected by tortuosity
        alphalh     = 0.1     , # longitudinal for horizontal flow 
        alphalv     = 0.1     , # longitudinal for vertical flow
        alphath     = 0.01    , # yy, zz isotropic transverse dispersivity
        alphatv     = 0.01    , # yy, zz isotropic transverse dispersivity
        stringid    = None    ,
        extension   = 'dsp'   ,
    ):

        # Call parent constructor
        ftype = self._ftype()
        if ( model.multipackage[ftype]['count'] == 0 ): 
            super().__init__(
                model,
                extension,
                ftype,
                model.multipackage[ftype]['unitnumber'],
                allowDuplicates=True,
            )
        else:
            # Pass the parent
            self._parent = model.multipackage[ftype]['instances'][0]._parent


        # modelkind
        if ( not isinstance( modelkind, str ) ):
            raise TypeError(
                f"{self.__class__.__name__}:"
                f" Invalid type for modelkind. It should be a str , but"
                f" {str(type(modelkind))} was given."
            )
        if modelkind is not None: 
            try:
                self.modelkindid = dispersionModel[modelkind.lower()].value
                self.modelkind   = modelkind.lower()
            except:
                raise ValueError(
                    f"{self.__class__.__name__}:"
                    f" Invalid dispersion modelkind {str(modelkind)}."
                    f" The allowed values are iso/isotropic or axi/axisymmetric."
                )
        else:
            self.modelkind   = 'iso'
            self.modelkindid = 0


        # basic preliminary health checks for dispersion params params
        if self.modelkindid == 0:
            if ( not isinstance(alphal,(int,float,np.ndarray,list)) ):
                raise TypeError(
                    f"{self.__class__.__name__}:"
                    f" Invalid type for alphal. It should be a real value int/float/list/np.ndarray."
                    f" {str(type(alphal))} was given."
                )
            if ( not isinstance(alphat,(int,float,np.ndarray,list)) ):
                raise TypeError(
                    f"{self.__class__.__name__}:"
                    f" Invalid type for alphat. It should be a real value int/float/list/np.ndarray."
                    f" {str(type(alphat))} was given."
                )
            if ( not isinstance(dmeff,(int,float,np.ndarray,list)) ):
                raise TypeError(
                    f"{self.__class__.__name__}:"
                    f" Invalid type for dmeff. It should be a real value int/float/list/np.ndarray."
                    f" {str(type(dmeff))} was given."
                )
        elif self.modelkindid == 1:
            if ( not isinstance(alphalh,(int,float,np.ndarray,list)) ):
                raise TypeError(
                    f"{self.__class__.__name__}:"
                    f" Invalid type for alphalh. It should be a real value int/float/list/np.ndarray."
                    f" {str(type(alphalh))} was given."
                )
            if ( not isinstance(alphalv,(int,float,np.ndarray,list)) ):
                raise TypeError(
                    f"{self.__class__.__name__}:"
                    f" Invalid type for alphalv. It should be a real value int/float/list/np.ndarray."
                    f" {str(type(alphalv))} was given."
                )
            if ( not isinstance(alphath,(int,float,np.ndarray,list)) ):
                raise TypeError(
                    f"{self.__class__.__name__}:"
                    f" Invalid type for alphath. It should be a real value int/float/list/np.ndarray."
                    f" {str(type(alphath))} was given."
                )
            if ( not isinstance(alphatv,(int,float,np.ndarray,list)) ):
                raise TypeError(
                    f"{self.__class__.__name__}:"
                    f" Invalid type for alphatv. It should be a real value int/float/list/np.ndarray."
                    f" {str(type(alphatv))} was given."
                )
            if ( not isinstance(dmeff,(int,float,np.ndarray,list)) ):
                raise TypeError(
                    f"{self.__class__.__name__}:"
                    f" Invalid type for dmeff. It should be a real value int/float/list/np.ndarray."
                    f" {str(type(dmeff))} was given."
                )


        # If no specific input format was given, 
        # infer from the type of parameters
        if ( inputformat is None ):
            # linear isotropic
            if self.modelkindid == 0:
                # uniform params
                if ( 
                    ( isinstance(alphal,(int,float)) ) and 
                    ( isinstance(alphat,(int,float)) ) and 
                    ( isinstance(dmeff ,(int,float)) )  
                ):
                    inputformat = 'uni'
                else:
                # non uniform params
                    if ( isinstance( alphal, (int,float) ) ):
                        alphal = [alphal]
                    if ( isinstance( alphat, (int,float) ) ):
                        alphat = [alphat]
                    if ( isinstance( dmeff, (int,float) ) ):
                        dmeff  = [dmeff]
                    if ( isinstance(alphal,list) ): 
                        alphal = np.array(alphal)
                    if ( isinstance(alphat,list) ): 
                        alphat = np.array(alphat)
                    if ( isinstance(dmeff,list) ): 
                        dmeff = np.array(dmeff)
                    if ( alphal.dtype not in [np.int32, np.int64, np.float32, np.float64] ):
                        raise TypeError(
                            f"{self.__class__.__name__}:"
                            f" Invalid type for alphal. It should be a real value, but"
                            f" {str(alphal.dtype)} was given."
                        )
                    if ( alphat.dtype not in [np.int32, np.int64, np.float32, np.float64] ):
                        raise TypeError(
                            f"{self.__class__.__name__}:"
                            f" Invalid type for alphat. It should be a real value, but"
                            f" {str(alphat.dtype)} was given."
                        )
                    if ( dmeff.dtype not in [np.int32, np.int64, np.float32, np.float64] ):
                        raise TypeError(
                            f"{self.__class__.__name__}:"
                            f" Invalid type for dmeff. It should be a real value, but"
                            f" {str(dmeff.dtype)} was given."
                        )
                    inputformat = 'dist'
            # linear axisymmetric
            elif self.modelkindid == 1:
                # uniform params
                if ( 
                    ( isinstance(alphalh,(int,float)) ) and 
                    ( isinstance(alphalv,(int,float)) ) and 
                    ( isinstance(alphath,(int,float)) ) and 
                    ( isinstance(alphatv,(int,float)) ) and 
                    ( isinstance(dmeff ,(int,float)) )  
                ):
                    inputformat = 'uni'
                else:
                # non uniform params
                    if ( isinstance( alphalh, (int,float) ) ):
                        alphalh = [alphalh]
                    if ( isinstance( alphalv, (int,float) ) ):
                        alphalv = [alphalv]
                    if ( isinstance( alphath, (int,float) ) ):
                        alphath = [alphath]
                    if ( isinstance( alphatv, (int,float) ) ):
                        alphatv = [alphatv]
                    if ( isinstance( dmeff, (int,float) ) ):
                        dmeff  = [dmeff]
                    if ( isinstance(alphalh,list) ): 
                        alphalh = np.array(alphalh)
                    if ( isinstance(alphalv,list) ): 
                        alphalv = np.array(alphalv)
                    if ( isinstance(alphath,list) ): 
                        alphath = np.array(alphath)
                    if ( isinstance(alphatv,list) ): 
                        alphatv = np.array(alphatv)
                    if ( isinstance(dmeff,list) ): 
                        dmeff = np.array(dmeff)
                    if ( alphalh.dtype not in [np.int32, np.int64, np.float32, np.float64] ):
                        raise TypeError(
                            f"{self.__class__.__name__}:"
                            f" Invalid type for alphalh. It should be a real value, but"
                            f" {str(alphalh.dtype)} was given."
                        )
                    if ( alphalv.dtype not in [np.int32, np.int64, np.float32, np.float64] ):
                        raise TypeError(
                            f"{self.__class__.__name__}:"
                            f" Invalid type for alphalv. It should be a real value, but"
                            f" {str(alphalv.dtype)} was given."
                        )
                    if ( alphath.dtype not in [np.int32, np.int64, np.float32, np.float64] ):
                        raise TypeError(
                            f"{self.__class__.__name__}:"
                            f" Invalid type for alphath. It should be a real value, but"
                            f" {str(alphath.dtype)} was given."
                        )
                    if ( alphatv.dtype not in [np.int32, np.int64, np.float32, np.float64] ):
                        raise TypeError(
                            f"{self.__class__.__name__}:"
                            f" Invalid type for alphatv. It should be a real value, but"
                            f" {str(alphatv.dtype)} was given."
                        )
                    if ( dmeff.dtype not in [np.int32, np.int64, np.float32, np.float64] ):
                        raise TypeError(
                            f"{self.__class__.__name__}:"
                            f" Invalid type for dmeff. It should be a real value, but"
                            f" {str(dmeff.dtype)} was given."
                        )
                    inputformat = 'dist'

        # inputformat
        if ( not isinstance( inputformat, str ) ):
            raise TypeError(
                f"{self.__class__.__name__}:"
                f" Invalid type for inputformat option {str(type(inputformat))}."
                f" It was expecting a string." 
            )
        try:
            self.inputformat = inputFormat[inputformat.lower()].value
        except:
            raise ValueError(
                f"{self.__class__.__name__}:"
                f" Invalid input format option {str(inputformat)}."
                f" The allowed values are uni/uniform or dis/distributed/u3d." 
            )

        # Assign params
        if self.inputformat == 0:
            # Spatially constant, uniform 
            if self.modelkindid == 0:
                # Validate types
                # alphal
                if ( not isinstance(alphal,(int,float)) ):
                    if ( isinstance(alphal,(list,np.ndarray)) and (len(alphal)==1)):
                        alphal = alphal[0]
                    else:
                        raise TypeError(
                            f"{self.__class__.__name__}:"
                            f" Invalid type for alphal. Uniform format was requested, but type"
                            f" {str(type(alphal))} was given."
                        )
                # alphat
                if ( not isinstance(alphat,(int,float)) ):
                    if ( isinstance(alphat,(list,np.ndarray)) and (len(alphat)==1)):
                        alphat = alphat[0]
                    else:
                        raise TypeError(
                            f"{self.__class__.__name__}:"
                            f" Invalid type for alphat. Uniform format was requested, but type"
                            f" {str(type(alphat))} was given."
                        )
                # dmeff
                if ( not isinstance(dmeff,(int,float)) ):
                    if ( isinstance(dmeff,(list,np.ndarray)) and (len(dmeff)==1)):
                        dmeff = dmeff[0]
                    else:
                        raise TypeError(
                            f"{self.__class__.__name__}:"
                            f" Invalid type for dmeff. Uniform format was requested, but type"
                            f" {str(type(dmeff))} was given."
                        )
                # Validate values
                if ( alphal < 0 ):
                    raise ValueError(
                        self.__class__.__name__ + ':'
                        + ' Inconsistent value of alphal ' +str(alphal)
                        + '. It should be greater or equal than zero.'
                    )
                if ( alphat < 0 ):
                    raise ValueError(
                        f"{self.__class__.__name__}:"
                        f" Inconsistent value of alphat {str(alphat)}."
                        f" It should be greater or equal than zero."
                    )
                if ( dmeff < 0 ):
                    raise ValueError(
                        f"{self.__class__.__name__}:"
                        f" Inconsistent value of dmeff {str(dmeff)}."
                        f" It should be greater or equal than zero."
                    )
                self.alphal = alphal 
                self.alphat = alphat
                self.dmeff  = dmeff

            elif self.modelkindid == 1:
                # linear axisymmetric
                # alphalh
                if ( not isinstance(alphalh,(int,float)) ):
                    if ( isinstance(alphalh,(list,np.ndarray)) and (len(alphalh)==1)):
                        alphalh = alphalh[0]
                    else:
                        raise TypeError(
                            f"{self.__class__.__name__}:"
                            f" Invalid type for alphalh. Uniform format was requested, but type"
                            f" {str(type(alphalh))} was given."
                        )
                # alphalv
                if ( not isinstance(alphalv,(int,float)) ):
                    if ( isinstance(alphalv,(list,np.ndarray)) and (len(alphalv)==1)):
                        alphalv = alphalv[0]
                    else:
                        raise TypeError(
                            f"{self.__class__.__name__}:"
                            f" Invalid type for alphalv. Uniform format was requested, but type"
                            f" {str(type(alphalv))} was given."
                        )
                # alphath
                if ( not isinstance(alphath,(int,float)) ):
                    if ( isinstance(alphath,(list,np.ndarray)) and (len(alphath)==1)):
                        alphath = alphath[0]
                    else:
                        raise TypeError(
                            f"{self.__class__.__name__}:"
                            f" Invalid type for alphath. Uniform format was requested, but type"
                            f" {str(type(alphath))} was given."
                        )
                # alphatv
                if ( not isinstance(alphatv,(int,float)) ):
                    if ( isinstance(alphatv,(list,np.ndarray)) and (len(alphatv)==1)):
                        alphatv = alphatv[0]
                    else:
                        raise TypeError(
                            f"{self.__class__.__name__}:"
                            f" Invalid type for alphatv. Uniform format was requested, but type"
                            f" {str(type(alphatv))} was given."
                        )
                # dmeff
                if ( not isinstance(dmeff,(int,float)) ):
                    if ( isinstance(dmeff,(list,np.ndarray)) and (len(dmeff)==1)):
                        dmeff = dmeff[0]
                    else:
                        raise TypeError(
                            f"{self.__class__.__name__}:"
                            f" Invalid type for dmeff. Uniform format was requested, but type"
                            f" {str(type(dmeff))} was given."
                        )
                # Validate values
                if ( alphalh < 0 ):
                    raise ValueError(
                        self.__class__.__name__ + ':'
                        + ' Inconsistent value of alphalh ' +str(alphalh)
                        + '. It should be greater or equal than zero.'
                    )
                if ( alphalv < 0 ):
                    raise ValueError(
                        self.__class__.__name__ + ':'
                        + ' Inconsistent value of alphalv ' +str(alphalv)
                        + '. It should be greater or equal than zero.'
                    )
                if ( alphath < 0 ):
                    raise ValueError(
                        f"{self.__class__.__name__}:"
                        f" Inconsistent value of alphath {str(alphath)}."
                        f" It should be greater or equal than zero."
                    )
                if ( alphatv < 0 ):
                    raise ValueError(
                        f"{self.__class__.__name__}:"
                        f" Inconsistent value of alphatv {str(alphatv)}."
                        f" It should be greater or equal than zero."
                    )
                if ( dmeff < 0 ):
                    raise ValueError(
                        f"{self.__class__.__name__}:"
                        f" Inconsistent value of dmeff {str(dmeff)}."
                        f" It should be greater or equal than zero."
                    )
                self.alphalh = alphalh
                self.alphalv = alphalv 
                self.alphath = alphath
                self.alphatv = alphatv
                self.dmeff   = dmeff
            else:
                raise NotImplementedError(
                        f"{self.__class__.__name__}:"
                        f" Dispersion model kind {str(self.modelkind)}"
                        f" is not implemented."
                    )
        elif self.inputformat == 1:
            # Distributed dispersion parameters 
            shape = model.shape
            if len(shape) == 3:
                shape3d = shape
            elif len(shape) == 2:
                shape3d = (shape[0], 1, shape[1])
            else:
                shape3d = (1, 1, shape[0])
            self.model = model
            self.shape3d = shape3d

            if self.modelkindid == 0:
                # linear isotropic
                try:
                    self.alphal = Util3d(
                        model,
                        shape3d,
                        np.float32,
                        alphal,
                        name="ALPHAL",
                        locat=self._parent.multipackage[ftype]['unitnumber'], 
                    )
                except:
                    raise Exception(
                        f"{self.__class__.__name__}:"
                        f" Error while initializing distributed variable alphal."
                        f" Is the input shape consistent with flow model dimensions ?"
                    )
                try:
                    self.alphat = Util3d(
                        model,
                        shape3d,
                        np.float32,
                        alphat,
                        name="ALPHAT",
                        locat=self._parent.multipackage[ftype]['unitnumber'], 
                    )
                except:
                    raise Exception(
                        f"{self.__class__.__name__}:"
                        f" Error while initializing distributed variable alphat."
                        f" Is the input shape consistent with flow model dimensions ?"
                    )
                try:
                    self.dmeff = Util3d(
                        model,
                        shape3d,
                        np.float32,
                        dmeff,
                        name="DMEFF",
                        locat=self._parent.multipackage[ftype]['unitnumber'], 
                    )
                except:
                    raise Exception(
                        f"{self.__class__.__name__}:"
                        f" Error while initializing distributed variable dmeff."
                        f" Is the input shape consistent with flow model dimensions ?"
                    )

                # Validate values
                if np.any( self.alphal.array < 0 ) : 
                    raise ValueError(
                        f"{self.__class__.__name__}:"
                        f" Inconsistent values for alphal. Negatives were found and"
                        f" it should be greater or equal than zero."
                    )
                if np.any( self.alphat.array < 0 ) : 
                    raise ValueError(
                        f"{self.__class__.__name__}:"
                        f" Inconsistent values for alphat. Negatives were found and"
                        f" it should be greater or equal than zero."
                    )
                if np.any( self.dmeff.array < 0 ) : 
                    raise ValueError(
                        f"{self.__class__.__name__}:"
                        f" Inconsistent values for dmeff. Negatives were found and"
                        f" it should be greater or equal than zero."
                    )
            elif self.modelkindid == 1:
                # linear axisymmetric
                # alphalh
                try:
                    self.alphalh = Util3d(
                        model,
                        shape3d,
                        np.float32,
                        alphalh,
                        name="ALPHALH",
                        locat=self._parent.multipackage[ftype]['unitnumber'], 
                    )
                except:
                    raise Exception(
                        f"{self.__class__.__name__}:"
                        f" Error while initializing distributed variable alphalh."
                        f" Is the input shape consistent with flow model dimensions ?"
                    )
                # alphalv
                try:
                    self.alphalv = Util3d(
                        model,
                        shape3d,
                        np.float32,
                        alphalv,
                        name="ALPHALV",
                        locat=self._parent.multipackage[ftype]['unitnumber'], 
                    )
                except:
                    raise Exception(
                        f"{self.__class__.__name__}:"
                        f" Error while initializing distributed variable alphalv."
                        f" Is the input shape consistent with flow model dimensions ?"
                    )
                # alphath
                try:
                    self.alphath = Util3d(
                        model,
                        shape3d,
                        np.float32,
                        alphath,
                        name="ALPHATH",
                        locat=self._parent.multipackage[ftype]['unitnumber'], 
                    )
                except:
                    raise Exception(
                        f"{self.__class__.__name__}:"
                        f" Error while initializing distributed variable alphath."
                        f" Is the input shape consistent with flow model dimensions ?"
                    )
                # alphatv
                try:
                    self.alphatv = Util3d(
                        model,
                        shape3d,
                        np.float32,
                        alphatv,
                        name="ALPHATV",
                        locat=self._parent.multipackage[ftype]['unitnumber'], 
                    )
                except:
                    raise Exception(
                        f"{self.__class__.__name__}:"
                        f" Error while initializing distributed variable alphatv."
                        f" Is the input shape consistent with flow model dimensions ?"
                    )
                # dmeff
                try:
                    self.dmeff = Util3d(
                        model,
                        shape3d,
                        np.float32,
                        dmeff,
                        name="DMEFF",
                        locat=self._parent.multipackage[ftype]['unitnumber'], 
                    )
                except:
                    raise Exception(
                        f"{self.__class__.__name__}:"
                        f" Error while initializing distributed variable dmeff."
                        f" Is the input shape consistent with flow model dimensions ?"
                    )

                # Validate values
                if np.any( self.alphalh.array < 0 ) : 
                    raise ValueError(
                        f"{self.__class__.__name__}:"
                        f" Inconsistent values for alphalh. Negatives were found and"
                        f" it should be greater or equal than zero."
                    )
                if np.any( self.alphalv.array < 0 ) : 
                    raise ValueError(
                        f"{self.__class__.__name__}:"
                        f" Inconsistent values for alphalv. Negatives were found and"
                        f" it should be greater or equal than zero."
                    )
                if np.any( self.alphath.array < 0 ) : 
                    raise ValueError(
                        f"{self.__class__.__name__}:"
                        f" Inconsistent values for alphath. Negatives were found and"
                        f" it should be greater or equal than zero."
                    )
                if np.any( self.alphatv.array < 0 ) : 
                    raise ValueError(
                        f"{self.__class__.__name__}:"
                        f" Inconsistent values for alphatv. Negatives were found and"
                        f" it should be greater or equal than zero."
                    )
                if np.any( self.dmeff.array < 0 ) : 
                    raise ValueError(
                        f"{self.__class__.__name__}:"
                        f" Inconsistent values for dmeff. Negatives were found and"
                        f" it should be greater or equal than zero."
                    )
            else:
                raise NotImplementedError(
                        f"{self.__class__.__name__}:"
                        f" Dispersion model kind {str(self.modelkind)}"
                        f" is not implemented."
                    )

        # String id 
        if (stringid is not None): 
            if ( not isinstance( stringid, str ) ): 
                raise TypeError(
                    f"{self.__class__.__name__}:"
                    f" Invalid type for stringid. It should be str, but"
                    f" {str(type(stringid))} was given."
                )
            self.stringid = stringid
        else:
            self.stringid = ftype+str(self.id)


        # Done
        return


    def write_file(self, check=False):
        """
        Write the package file

        Parameters
        ----------
        check : boolean
            Check package data for common errors. (default False)
        Returns
        -------
        None
        """

        # Shortcut to multipackage dict
        mpkg = self.parent.multipackage[self._ftype()]
        
        # Write
        with open( mpkg['instances'][0].fn_path, 'w' ) as f:
        
            # Write how many instances...
            f.write(f"{mpkg['count']}\n")

            # ...and loop over them
            for ins in mpkg['instances']:
                
                # Write id
                f.write(f"{ins.stringid}\n")

                # Write modelkindid and input format
                f.write(f"{ins.modelkindid}    {ins.inputformat}\n")
               
                # Write dispersion model parameters
                if ins.modelkindid == 0:
                # linear isotropic
                    if ins.inputformat == 0:
                        # Uniform
                        f.write(f"{ins.dmeff :.10f}\n")  
                        f.write(f"{ins.alphal:.10f}\n")  
                        f.write(f"{ins.alphat:.10f}\n")  
                    elif ins.inputformat == 1:
                        # Distributed
                        f.write(ins.dmeff.get_file_entry())
                        f.write(ins.alphal.get_file_entry())
                        f.write(ins.alphat.get_file_entry())
                if ins.modelkindid == 1:
                # linear axisymmetric
                    if ins.inputformat == 0:
                        # Uniform
                        f.write(f"{ins.dmeff :.10f}\n")  
                        f.write(f"{ins.alphalh:.10f}\n")  
                        f.write(f"{ins.alphalv:.10f}\n")  
                        f.write(f"{ins.alphath:.10f}\n")  
                        f.write(f"{ins.alphatv:.10f}\n")  
                    elif ins.inputformat == 1:
                        # Distributed
                        f.write(ins.dmeff.get_file_entry())
                        f.write(ins.alphalh.get_file_entry())
                        f.write(ins.alphalv.get_file_entry())
                        f.write(ins.alphath.get_file_entry())
                        f.write(ins.alphatv.get_file_entry())
       
        # Done 
        return



'''
Configuration of MODPATH-RW observations 
'''

# python
import numpy as np
from enum import Enum

# flopy
from flopy.pakbase import Package
from flopy.utils import Util3d
from flopy.discretization import StructuredGrid

# local 
from .utils import multipackage


class observationKindOption(Enum): 
    '''
    Enumerate formats for observation kinds
    '''
    res      = 0
    resident = 0
    flux     = 1
    sink     = 1


class histogramBinOption(Enum): 
    '''
    Enumerate formats for observation kinds
    '''
    scott    = 0
    fd       = 1
    given    = 2
    ts       = 3


class ModpathRWObs( Package ):
    """
    MODPATH-RW Observation Package Class.

    Parameters
    ----------
    model : model object
        The model object (of type :class: ModpathRW) to which
        this package will be added.
    kind  : int or str
        The kind of observation. Allowed values are:
        * 0 or ('res','resident'): observation of resident concentrations, any cell.
        * 1 or ('flux','sink')   : observation of flux concentrations. For consistency, 
             it should be applied to cells with sink flow and 
             configure the simulation with weaksinkoption=stop_at.
    cellinputoption : int
        The format in which observation cells will be given. Allowed values are:
        * 0: cells are given as a list of individual cells.
        * 1: cells read with free-format input.
    cells : list, np.array
        The list of cells to be considered for the observation.
        * If cellinputoption == 0, then this parameter contain the list of cell ids
        * If cellinputoption == 1, this should an array with the dimensions
          of the flow model. In this last case, array entries containing ones
          are included as part of the observation.
    structured : bool 
        Indicates the format of cell ids in the cells param. If True, then 
        cells are in format (lay, row, col), and if False these are the 
        linear cell numbers.
    outputoption : int
        Defines the information stored in the output file. Allowed values are:
        * 0: output file contains the source particle records
        * 1: postprocessed observation in output file (timeseries) 
             and a second file with source records
        * 2: only one file with the postprocessed observation (timeseries)
    postprocessoption : int
        The kind of postprocess to be applied to the source observation records. 
        Allowed values are:
        * 0: timeseries with histogram reconstruction
        * 1: timeseries with both histogram and smoothed reconstruction
    reconstructionoptions : bool
        Enable interpretation of reconstruction parameters. 
        * False: timeseries reconstruction with default parameters. 
        * True : timeseries reconstruction with parameters provided by the user.
    noptloops : int
        The number of bandwidth optimization loops. It should be a positive integer.
    convergence : float
        The limit relative change of the variables monitored during bandwidth optimization 
        to determine the convergence of reconstruction. It should be positive.
    initialsmoothingformat : int
        Determines the protocol for selection of the initial kernel size from where to
        begin the optimization for bandwidth selection. Allowed values are: 
          * 0: selects initial bandwidth from the expression of Silverman (1986) 
               for Gaussian distributions. It is based on the standard deviation of
               the particle coordinates, and the number of particles. 
          * 1: initial bandwidth as a factor multiplying the characteristic cell size.
               It will use the parameter binsizefactor to perform the scaling.
    binsizefactor : float
        Scaling factor employed for determining the initial kernel size from where to start
        the bandwidth optimization. Amplifies the bin size. In timeseries reconstruction, 
        the bin size is time step defined for the timeseries simulation. 
    effectiveweightformat : int
        Determines the protocol for handling the reconstruction of particle distributions 
        with different weights. Allowed values are: 
          * 0: computes a unique effective number of particles and determines a 
               domain-effective particle weight (Kish, 1965,1992), which is used to
               transform the mass histogram into an equivalent particle distribution 
               used for the bandwidth selection.
          * 1: similar to the previous alternative, but the employed characteristic 
               particle weight is the average over all particles. 
          * 2: Bandwidth selection is performed taking into account only the particle 
               coordinates and a final density estimation is performed over the mass 
               histogram with the obtain distribution of kernel sizes.
          * 3: An effective number of particles is computed for each histogram cell and 
               is employed to transform the mass distribution into an effective particle 
               distribution used for the bandwidth optimization. A final reconstruction
               stage is performed considering the obtained distribution of kernel sizes.
    histogramoptions : bool 
        Enable interpretation of histogram parameters. The histogram is the base over 
        which smoothed reconstruction is performed. 
        * False: histogram options handled internally with default parameters, not written. 
        * True : interpret histogram options, written to package file.
    binoption : int or str
        Selection of the histogram bin size. Allowed values are: 
          * 0 or 'scott': applies the automatic selection for Gaussian distributions.
          * 1 or 'fd'   : applies the automatic selection based on the inter quartile range.
          * 2 or 'given': user provided value in binparam.
          * 3 or 'ts'   : infered from the timeseries, assumed uniform time step.
    binparam : float
        Auxiliary parameter controlling the selection of the histogram bin size. 
        If binoption is 0 or 1, binparam is fraction (0,1] multiplying the automatic value. 
        If binoption is 2, binparam is the bin size. 
        If binoption is 3, binparam is not read.
    timestepout : float
        Value determining the time discretization to write the timeseries data. If given, 
        interpolates the values obtained from the model to the given resolution. 
        Interpolation is linear and performed only within the limits of the real data. 
    filename : str
        The output file where the observation is written, the name 
        without the extension. If None is given, will employ stringid. 
        * If outputoption == 0, this file contains the observation records
        * If outputoption == 1, this file contains the postprocessed 
          observation and a second file whose name is rec+filename
          contains the observations records. 
        * If outputoption == 2, this file contains the postprocessed 
          observation timeseries
    stringid : str
        An identifier for the observation. By default assign the concatenation
        of package ftype and an integer id. 
    extension : str
        The extension for the observation file. By default 'obs' 
    """


    @staticmethod
    def _ftype():
        return 'OBS'


    @multipackage
    def __init__(
        self,
        model,
        kind                   = 0,
        cellinputoption        = 0,
        cells                  = None,
        structured             = None, 
        outputoption           = 2,
        postprocessoption      = 1,
        reconstructionoptions  = False, 
        noptloops              = 10, 
        convergence            = 0.01,
        initialsmoothingformat = 0,
        binsizefactor          = 3.0,
        effectiveweightformat  = 0,
        histogramoptions       = False, 
        binoption              = 1, 
        binparam               = None,
        timestepout            = None, 
        filename               = None, 
        stringid               = None,
        extension              = 'obs',
    ):

        # Call parent constructor
        ftype = self._ftype()
        if ( model.multipackage[ftype]['count'] == 0 ): 
            super().__init__(
                model,
                extension,
                ftype,
                model.multipackage[ftype]['unitnumber']
            )
        else:
            # Pass the parent
            self._parent = model.multipackage[ftype]['instances'][0]._parent


        # Observation kind
        if ( not isinstance( kind, (int,str) ) ):
            raise TypeError(
                f"{self.__class__.__name__}:"
                f" Invalid type for observation kind."
                f" It can be specified as int or str, but {str(type(kind))}"
                f" was given." 
            )
        if ( isinstance( kind, int ) ):
            if ( kind not in [0,1] ): 
                raise ValueError(
                    f"{self.__class__.__name__}:"
                    f" Invalid kind option {str(kind)}."
                    f" Allowed values are 0 (resident) or 1 (flux)."
                )
            self.kind = kind
            if self.kind == 0: 
                self.stringkind = 'RESIDENT'
            elif self.kind == 1:
                self.stringkind = 'FLUX'
        elif ( isinstance( kind, str ) ):
            try:
                self.kind = observationKindOption[kind.lower()].value
                if self.kind == 0: 
                    self.stringkind = 'RESIDENT'
                elif self.kind == 1:
                    self.stringkind = 'FLUX'
            except:
                raise ValueError(
                    f"{self.__class__.__name__}:"
                    f" Invalid observation kind {str(kind)}."
                    f" The allowed values are resident or flux."
                )

        # output option
        if ( outputoption not in [0,1,2] ):
            raise ValueError(
                f"{self.__class__.__name__}:"
                f" Invalid output option {str(kind)}."
                f" Allowed values are 0 (only obs records),"
                f" 1 (records and postprocess) or 2 (only postprocess)."
            )
        self.outputoption = outputoption

        # postprocess option
        if ( postprocessoption not in [0,1] ):
            raise ValueError(
                f"{self.__class__.__name__}:"
                f" Invalid postprocess option {str(kind)}."
                f" Allowed values are 0 (only histogram) or 1"
                f" (histogram and smoothed reconstruction)."
            )
        self.postprocessoption = postprocessoption

        # cellinputoption 
        if ( cellinputoption not in [0,1] ):
            raise ValueError(
                f"{self.__class__.__name__}:"
                f" Invalid cellinputoption option {str(cellinputoption)}"
                f" Allowed values are 0 (list of cellids) or 1 (modelgrid array)."
            )
        self.cellinputoption = cellinputoption

        if self.cellinputoption == 0:

            if ( cells is None ):
                raise ValueError(
                    f"{self.__class__.__name__}:"
                    f" Invalid cells specification. It was expecting a list"
                    f" of cell ids, but None was given."
                )

            # Infer structured parameter from 
            # the modelgrid type if it remained with 
            # default None
            if ( structured is None ):
                structured = False
                if ( isinstance( self._parent.flowmodel.modelgrid, StructuredGrid ) ):
                    # If grid is structured, infers True
                    structured = True

            if ( not isinstance( structured, bool ) ): 
                raise TypeError(
                    f"{self.__class__.__name__}:"
                    f" Invalid type for structured parameter."
                    f" It should be boolean, but"
                    f" {str(type(structured))} was given."
                )

            # Cell format
            if structured:
                self.cellformat = 0
            else:
                self.cellformat = 1
            self.structured = structured
            
            # Some validation with the modelgrid
            if (
                (self.structured) and
                not isinstance( self._parent.flowmodel.modelgrid, StructuredGrid )
            ):
                raise Exception(
                    f"{self.__class__.__name__}:"
                    f" Invalid cellinputoption and structured parameters"
                    f" for the modelgrid of type {str(type(self._parent.flowmodel.modelgrid))}."
                    f" Needs to be StructuredGrid in order to specify cells as (lay,row,col)."
                )

            # Write as a list of cellids
            if cells is None:
                self.cells = []
            else:
                if not isinstance( cells, (list,np.ndarray,tuple) ):
                    raise TypeError(
                        f"{self.__class__.__name__}:"
                        f" Invalid input cells type."
                        f" It should be list, tuple of np.ndarray."
                        f" {str(type(cells))} was given."
                    )
                # Transform to numpy array
                cells = np.array( cells )
                
                if self.structured: 
                    # The case of a single list/tuple with cell id
                    if ( (len(cells.shape)==1) ): 
                        if ( len(cells)==3 ):
                            cells = np.expand_dims( cells, axis=0 )
                        else:
                            raise ValueError(
                                f"{self.__class__.__name__}:"
                                f" Invalid cells specification. While using the"
                                f" structured flag for cellinputoption = 0, the cells"
                                f" should be specified as a list of cell ids following"
                                f" the format (lay,row,col)."
                            )
                    elif ( len(cells.shape)==2 ):
                        if ( not ( cells.shape[1] == 3 ) ): 
                            raise ValueError(
                                f"{self.__class__.__name__}:"
                                f" Invalid cells specification. While using the"
                                f" structured flag for cellinputoption = 0, the cells"
                                f" should be specified as a list of cell ids following"
                                f" the format (lay,row,col)."
                            )
                    else:
                        raise ValueError(
                            f"{self.__class__.__name__}:"
                            f" Invalid cells specification. While using the"
                            f" structured flag for cellinputoption = 0, the cells"
                            f" should be specified as a list of cell ids following"
                            f" the format (lay,row,col)."
                        )
                else:
                    # For unstructured, ask for a one-dimensional array
                    if ( (len(cells.shape)!=1) ): 
                        raise ValueError(
                            f"{self.__class__.__name__}:"
                            f" Invalid cells specification. While using"
                            f" the unstructured format for cellinputoption=0,"
                            f" cell ids should be given as linear indexes format."
                        )
                # Validate dtype
                if ( cells.dtype not in [np.int32,np.int64,int] ):
                    raise TypeError(
                        f"{self.__class__.__name__}:"
                        f" Invalid type for cells specification." 
                        f" Cell ids should be of integer type, but"
                        f" {str(cells.dtype)} was given."
                    )

                # To the class
                self.cells = cells

            # If no cells, stop. 
            if len(self.cells) == 0:
                raise ValueError( 
                    f"{self.__class__.__name__}:"
                    f" No cells were given for the observation."
                )

        elif self.cellinputoption == 1: 
            # Write obs cells as 3D array
            shape = model.shape
            if len(shape) == 3:
                shape3d = shape
            elif len(shape) == 2:
                shape3d = (shape[0], 1, shape[1])
            else:
                shape3d = (1, 1, shape[0])
            self.model   = model
            self.shape3d = shape3d

            if ( cells is None ):
                raise ValueError( 
                    f"{self.__class__.__name__}:"
                    f" No cells were given for the observation."
                )
            else:
                if not isinstance( cells, (int,list,np.ndarray) ):
                    raise TypeError(
                        f"{self.__class__.__name__}:"
                        f" Invalid type for input cells. It should be list or np.ndarray."
                        f" {str(type(cells))} was given."
                    )
                cells  = np.array(cells)
                ucells = np.unique(cells)
                for uc in ucells: 
                    if uc not in [0,1]:
                        raise ValueError( 
                            f"{self.__class__.__name__}:"
                            f" Invalid values for cells specification."
                            f" While using cellinputoption=1, the cells specification"
                            f" should only contain values 0 or 1."
                            f" The value {str(uc)} was found."
                        )

                try:
                    self.cells = Util3d(
                        model,
                        shape3d,
                        np.int32,
                        cells,
                        name="OBSCELLS",
                        locat=self._parent.multipackage[ftype]['unitnumber'], 
                    )
                except:
                    raise Exception(
                        f"{self.__class__.__name__}:"
                        f" Error while initializing distributed variable cells."
                        f" Is the input shape consistent with flow model dimensions ?"
                    )

        if ( not isinstance( reconstructionoptions,bool ) ):
            raise TypeError(
                f"{self.__class__.__name__}:"
                f" Invalid type for reconstructionoptions. It should be bool, but"
                f" {str(type(reconstructionoptions))} was given."
            )
        if reconstructionoptions:
            self.reconstructionoptions = 1
        else:
            self.reconstructionoptions = 0

        # process reconstruction options
        if reconstructionoptions:

            # noptloops
            if ( not isinstance( noptloops,int ) ):
                raise TypeError(
                    f"{self.__class__.__name__}:"
                    f" Invalid type for noptloops. It should be int, but"
                    f" {str(type(noptloops))} was given."
                )
            if ( noptloops < 0 ):
                raise ValueError(
                    f"{self.__class__.__name__}:"
                    f" Invalid value for noptloops. It should be positive or zero."
                )
            self.noptloops = noptloops

            # convergence
            if ( not isinstance( convergence, (int,float) ) ):
                raise TypeError(
                    f"{self.__class__.__name__}:"
                    f" Invalid type for convergence. It should be float, but"
                    f" {str(type(convergence))} was given."
                )
            if ( convergence <= 0 ):
                raise ValueError(
                    f"{self.__class__.__name__}:"
                    f" Invalid value for convergence. It should be a positive real number."
                )
            self.convergence = convergence

            # initialsmoothingformat
            if ( not isinstance( initialsmoothingformat,int ) ):
                raise TypeError(
                    f"{self.__class__.__name__}:"
                    f" Invalid type for initialsmoothingformat. It should be int, but"
                    f" {str(type(initialsmoothingformat))} was given."
                )
            if initialsmoothingformat not in [0,1]: 
                raise ValueError(
                    f"{self.__class__.__name__}:"
                    f" Invalid value for initialsmoothinformat. Allowed values are"
                    f" 0 (automatic for Gaussian) or 1 (factor multiplying bin size)."
                )
            self.initialsmoothingformat = initialsmoothingformat

            # binsizefactor
            if ( not isinstance( binsizefactor, (int,float) ) ):
                raise TypeError(
                    f"{self.__class__.__name__}:"
                    f" Invalid type for binsizefactor. It should be float, but"
                    f" {str(type(binsizefactor))} was given."
                )
            if ( binsizefactor <= 0 ):
                raise ValueError(
                    f"{self.__class__.__name__}:"
                    f" Invalid value for binsizefactor. It should be a positive real number."
                )
            self.binsizefactor = binsizefactor

            # effectiveweightformat
            if ( not isinstance( effectiveweightformat,int ) ):
                raise TypeError(
                    f"{self.__class__.__name__}:"
                    f" Invalid type for effectiveweightformat. It should be int, but"
                    f" {str(type(effectiveweightformat))} was given."
                )
            if effectiveweightformat not in [0,1,2,3]: 
                raise ValueError(
                    f"{self.__class__.__name__}:" 
                    f" Invalid value for effectiveweightformat. Determines transformation"
                    f" to an equivalen count of particles. Allowed values are"
                    f" 0 (unique effective mass), 1 (scale by averaged mass),"
                    f" 2 (real histogram) or 3 (local effective histogram)."
                    f" {str(effectiveweightformat)} was given."
                )
            self.effectiveweightformat = effectiveweightformat


        if ( not isinstance( histogramoptions,bool ) ):
            raise TypeError(
                f"{self.__class__.__name__}:"
                f" Invalid type for histogramoptions. It should be bool, but"
                f" {str(type(histogramoptions))} was given."
            )
        if histogramoptions:
            self.histogramoptions = 1
        else:
            self.histogramoptions = 0

        if histogramoptions: 

            # binoption
            if ( not isinstance( binoption,int ) ):
                raise TypeError(
                    f"{self.__class__.__name__}:"
                    f" Invalid type for binoption. It should be int, but"
                    f" {str(type(binoption))} was given."
                )
            if binoption not in [0,1,2,3]: 
                raise ValueError(
                    f"{self.__class__.__name__}:" 
                    f" Invalid value for binoption. Determines selection"
                    f" of histogram bin size. Allowed values are"
                    f" 0 (based on Scott's rule), 1 (based on Freedman-Diaconis rule),"
                    f" 2 (given by user) or 3 (infered from timepointdata)."
                    f" {str(binoption)} was given."
                )
            self.binoption = binoption

            # binparam
            if binparam is None: 

                if ( self.binoption in [0,1] ):
                    # Set a default 
                    self.binparam = 0.75

                if ( self.binoption == 2 ):
                    # Needs to be requested
                    raise ValueError(
                        f"{self.__class__.__name__}:"
                        f" Invalid value for binparam. None was given."
                    )

            else:

                if ( not isinstance( binparam, (int,float) ) ):
                    raise TypeError(
                        f"{self.__class__.__name__}:"
                        f" Invalid type for binparam. It should be float, but"
                        f" {str(type(binparam))} was given."
                    )

                if ( self.binoption in [0,1] ):
                    if ( ( binparam <= 0 ) or ( binparam > 1 ) ):
                        raise ValueError(
                            f"{self.__class__.__name__}:"
                            f" Invalid value for binparam. It should be a fraction."
                        )
                    self.binparam = binparam

                if ( self.binoption == 2 ):
                    if ( binparam <= 0 ):
                        raise ValueError(
                            f"{self.__class__.__name__}:"
                            f" Invalid value for binparam. It should be a positive real number."
                        )
                    self.binparam = binparam

            # timestepout
            if timestepout is not None: 
                if ( not isinstance( timestepout, (int,float) ) ):
                    raise TypeError(
                        f"{self.__class__.__name__}:"
                        f" Invalid type for timestepout. It should be float, but"
                        f" {str(type(timestepout))} was given."
                    )
                if ( timestepout <= 0 ):
                    raise ValueError(
                        f"{self.__class__.__name__}:"
                        f" Invalid value for timestepout. It should be a positive real number."
                    )
                self.timestepout = timestepout
            else:
                self.timestepout = None


        # String id
        if (stringid is not None):
            if ( not isinstance( stringid, str ) ): 
                raise TypeError(
                    f"{self.__class__.__name__}:"
                    f" Invalid type for stringid. It should be str, but"
                    f" {str(type(stringid))} was given."
                )
            self.stringid = stringid
        else:
            self.stringid = ftype+str(self.id) 

        # filename   
        if ( filename is None ):
            # Assign by default the string id, in lower case
            filename = self.stringid.lower()
        else:
            # If a value was given, validate
            if ( not isinstance( filename, str ) ): 
                raise TypeError(
                    f"{self.__class__.__name__}:"
                    f" Invalid type for filename. It should be str, but"
                    f" {str(type(filename))} was given."
                )
        # Some validation for extension
        if ( not isinstance( extension, str ) ): 
            raise TypeError(
                f"{self.__class__.__name__}:"
                f" Invalid type for extension. It should be str, but"
                f" {str(type(extension))} was given."
            )
        # Filename for this observation
        self.filename = f"{filename}.{extension}"

        # Done
        return


    def write_file(self, check=False):
        """
        Write the package file
        Parameters
        ----------
        check : boolean
            Check package data for common errors. (default False)
        Returns
        -------
        None
        """

        # Shortcut to multipackage dict
        mpkg = self.parent.multipackage[self._ftype()]
        
        # Write
        with open( mpkg['instances'][0].fn_path, 'w' ) as f:

            # Write how many instances...
            f.write(f"{mpkg['count']}\n")

            # ...and loop over them
            for ins in mpkg['instances']:

                # Write obs stringid
                f.write(f"{ins.stringid}\n")

                # Write obs stringkind
                f.write(f"{ins.stringkind}\n")

                # Write obs filename
                f.write(f"{ins.filename}\n")

                # For sink observations
                if ins.kind == 1:
                    # Write the observation options 
                    f.write(f"{ins.outputoption}  {ins.postprocessoption}  {ins.histogramoptions}  {ins.reconstructionoptions}\n")
                else:
                    # Write the observation options 
                    f.write(f"{ins.outputoption}  {ins.postprocessoption}  {ins.reconstructionoptions}\n")

                # write the histogram data
                if ( (ins.kind==1) and (ins.histogramoptions==1) ):
                    if ( ins.binoption==3 ): 
                        f.write(f"{ins.binoption}\n")
                    else:
                        if ( ins.timestepout is None ): 
                            f.write(f"{ins.binoption}  {ins.binparam:.10f}\n")
                        else:
                            f.write(f"{ins.binoption}  {ins.binparam:.10f}  {ins.timestepout:.10f}\n")

                if ins.reconstructionoptions == 1:
                    # Write noptloops and convergence
                    f.write(f"{ins.noptloops}   {ins.convergence}\n")

                    if ins.initialsmoothingformat == 1:
                        # Write initialsmoothingformat and binsizefactor
                        f.write(f"{ins.initialsmoothingformat}   {ins.binsizefactor}\n")
                    else:
                        # Write initialsmoothingformat
                        f.write(f"{ins.initialsmoothingformat}\n")

                    # Write effectiveweightformat
                    f.write(f"{ins.effectiveweightformat}\n")



                # Write the celloption params
                if ins.cellinputoption == 0:
                    # The flag and specs
                    f.write(f"{ins.cellinputoption}  {len(ins.cells)}  {ins.cellformat} \n")

                    # And the list
                    fmts = []
                    if ins.structured:
                        fmts.append("{:9d}") # lay
                        fmts.append("{:9d}") # row
                        fmts.append("{:9d}") # col
                    else:
                        fmts.append("{:9d}") # cellid
                    fmt = " " + " ".join(fmts) + "\n"
                    for oc in ins.cells:
                        woc = np.array(oc).astype(np.int32)+1 # Correct the zero-based indexes
                        if ins.structured:
                            f.write(fmt.format(*woc))
                        else:
                            f.write(fmt.format(woc))

                elif ins.cellinputoption == 1:

                    # The flag
                    f.write(f"{ins.cellinputoption}\n")

                    # And the array
                    f.write(ins.cells.get_file_entry())


        # Done
        return


    def get_output(self):
        '''
        Load the observation output file into a rec array

        Note: the current implementation loads data for the cases
              where the time vector is different for each observation, 
              that is, the bin size was selected automatically and without
              interpolation (data appended vertically).
        '''
        import os 
        from flopy.utils.flopy_io import loadtxt
        # Shouldn't this be a case to handle at loadtxt ?
        try:
            import pandas as pd
            use_pandas = True
        except ImportError:
            use_pandas = False
        

        if self.stringkind == 'RESIDENT':
            raise NotImplementedError(
                    f"{self.__class__.__name__}:get_output:"
                    f" Output loading is not yet implemented for observations of kind RESIDENT."
                )

        if self.stringkind == 'FLUX':
            if self.outputoption in [1,2]:
                # Loading structure at this point only for btc's
                # with specific time vector
                # This needs some review: in simulations 
                # with adaptgridtocoords = true (an endpoint for example)
                # the time vector can also be different for each species, 
                # although interpolation was applied.
                if (
                    ( self.histogramoptions==1 ) and (
                        ( self.binoption in [2,3] ) 
                        #or ( self.timestepout is not None )
                    )
                ):
                    raise NotImplementedError(
                            f"{self.__class__.__name__}:get_output:"
                            f" Output loading is not yet implemented for observations of type FLUX "
                            f" with the same time step for all species."
                        )

                # continue
                dtypelist = [
                        ("speciesid" , np.int32  ),
                        ("tid"       , np.int32  ),
                        ("time"      , np.float32),
                        ("qsink"     , np.float32),
                        ("chist"     , np.float32),
                    ]
                # extend list if postprocessed
                if self.postprocessoption == 1:
                    dtypelist.append( ("cgpkde"    , np.float32) )

                # load
                recdata = loadtxt(
                    os.path.join( self._parent.model_ws, self.filename ),
                    dtype=np.dtype(dtypelist),
                    skiprows=0,
                    use_pandas=use_pandas
                )
                
                # To python zero-based indexation
                recdata['speciesid'] = recdata['speciesid'] - 1

                # done
                return recdata

            # for observations with only records
            else:
                raise NotImplementedError(
                        f"{self.__class__.__name__}:get_output:"
                        f" Output loading is not yet implemented for observations of type FLUX "
                        f" storing only records."
                    )



'''
Configuration of MODPATH-RW species
'''

# python
import numpy as np

# flopy
from flopy.pakbase import Package

# local 
from .utils import multipackage
from .mprwdsp import ModpathRWDsp


class ModpathRWSpc( Package ):
    """
    MODPATH-RW Species Package Class.

    Parameters
    ----------
    model : model object
        The model object (of type :class: ModpathRW) to which
        this package will be added.
    dispersion: ModpathRWDsp object (required)
        Relates a dispersion model to this solute. Is a required foreign key. 
    pgroups : list, np.ndarray [int]
        Particle groups related to this specie. These are interpreted 
        only if particlesmassoption != 2. These ids are zero based 
        (while writing is corrected to one based) and are related to 
        the order in which the particle groups are specified to the program.
        In case the simulation is configured with particlesmassoption == 2, 
        then the particle groups associated to a specie are interpreted 
        (internally) from the specified solute ids for each particle group.
    stringid : str
        String identifier for this species. By default concatenates 
        package ftype and integer id.
    extension : str
        The extension for the species file, by default spc. 
    """


    @staticmethod
    def _ftype():
        return 'SPC'


    @multipackage
    def __init__(
        self,
        model,
        dispersion, # it's like a foreign key
        pgroups    = None,
        stringid   = None,
        extension  = 'spc',
    ):

        # Call parent constructor
        ftype = self._ftype()
        if ( model.multipackage[ftype]['count'] == 0 ): 
            super().__init__(
                model,
                extension,
                ftype,
                model.multipackage[ftype]['unitnumber'],
                allowDuplicates=True,
            )
        else:
            # Pass the parent
            self._parent = model.multipackage[ftype]['instances'][0]._parent

        # Assign dispersion
        if not isinstance( dispersion, ModpathRWDsp ):
            raise TypeError(
                f"{self.__class__.__name__}:"
                f" Invalid dispersion parameter of type {str(type(dispersion))}."
                f" It should be type ModpathRWDsp."
            )
        self.dispersion = dispersion

        # Will be written only if particlesmassoption != 2
        if pgroups is not None:
            if not isinstance(pgroups, (int,list,np.ndarray)):
                raise TypeError(
                    f"{self.__class__.__name__}:"
                    f" Invalid pgroups parameter of type {str(type(pgroups))}."
                    f" It should be of type int, list or np.ndarray."
                )
            if isinstance( pgroups, list ):
                pgroups = np.array( pgroups )
            if isinstance( pgroups, int ):
                pgroups = np.array( [pgroups] )
            # Requires some more sanity checks to verify integer id's.
            for pg in pgroups: 
                if ( not isinstance( pg, (int,np.int32,np.int64) ) ):
                    raise TypeError(
                        f"{self.__class__.__name__}:"
                        f" Invalid pgroups parameter of type {str(type(pg))}."
                        f" It should be an integer pgroup id."
                    )

        # Might be worth considering a verification versus 
        # specified particle groups, which would require
        # checking the existence of pgroups at the simulation pkg,
        # or valid IC/SRC packages.
        self.pgroups = pgroups
      
        # String id 
        if (stringid is not None): 
            if ( not isinstance( stringid, str ) ): 
                raise TypeError(
                    f"{self.__class__.__name__}:"
                    f" Invalid type for stringid. It should be str, but"
                    f" {str(type(stringid))} was given."
                )
            self.stringid = stringid
        else:
            self.stringid = ftype+str(self.id)

        # Done
        return


    def write_file(self, check=False):
        """
        Write the package file
        Parameters
        ----------
        check : boolean
            Check package data for common errors. (default False)
        Returns
        -------
        None
        """

        # Shortcut to multipackage dict
        mpkg = self.parent.multipackage[self._ftype()]
        
        # Write
        with open( mpkg['instances'][0].fn_path, 'w' ) as f:

            # Write how many instances...
            f.write(f"{mpkg['count']}\n")

            # ...and loop over them
            for ins in mpkg['instances']:

                # Write stringid
                f.write(f"{ins.stringid}\n")

                # Need to write the related groups 
                # only if particlesmassoption not equal 2
                # If particlesmassoption equal 2, then 
                # soluteId is infered from given particlegroups

                # BaseModel has a weird getattr method and if
                # for any reason the program is trying to write
                # a model without a defined simulation package, 
                # particlesmassoption is undefined and throws 
                # an unclear message obtained from the overloaded
                # __getattr__
                if ( not hasattr(self.parent,'particlesmassoption') ):
                    raise Exception( 
                        f"{self.__class__.__name__}:"
                        f" The particlesmassoption was not found in parent package."
                        f" Did you define a ModpathRWSim package ?"
                    )
                if ( self.parent.particlesmassoption != 2 ):
                    if ( 
                        ( ins.pgroups is not None ) and
                        ( len(ins.pgroups) > 0 )
                    ):
                        # Write the number of groups
                        f.write(f"{len(ins.pgroups)}\n")

                        for idpg, pg in enumerate(ins.pgroups):
                            # Write the pgroup id in the list of pgroups
                            # Notice the plus one.
                            f.write(f"{pg+1}\n")
                    else:
                        raise ValueError(
                            f"{self.__class__.__name__}:"
                            f" Invalid pgroups parameter of type {str(type(ins.pgroups))}."
                            f" While using particlesmassoption != 2, the ModpathRWSpc package"
                            f" requires the list of pgroups indexes related to this species."
                        )
                if ( not hasattr(self.parent,'speciesdispersionoption') ):
                    raise Exception( 
                        f"{self.__class__.__name__}:"
                        f" The speciesdispersionoption was not found in parent package." 
                        f" Did you define a ModpathRWSim package ?"
                    )
                # Write dispersion "foreign key", only if
                # speciesdispersionoption == 1
                if self.parent.speciesdispersionoption == 1:
                    f.write(f"{ins.dispersion.stringid}\n")


        # Done
        return



'''
Configuration of MODPATH-RW initial conditions
'''

# python
import numpy as np
from enum import Enum

# flopy
from flopy.pakbase import Package
from flopy.utils import Util3d

# local 
from .utils import multipackage


class icKindOption(Enum): 
    '''
    Enumerate formats for timeseries output option
    '''
    globalmass = 0
    localmass  = 1


class particlesDistributionOption(Enum): 
    '''
    Enumerate formats for timeseries output option
    '''
    equ         = 0 
    equispaced  = 0
    qra         = 1 
    quasirandom = 1
    ran         = 2
    random      = 2


class ModpathRWIc( Package ):
    """
    MODPATH-RW Initial Condition Package Class.

    Parameters
    ----------
    model : model object
        The model object (of type :class: ModpathRW) to which
        this package will be added.
    kind  : int or str 
        The kind of initial condition specification. All formats load concentration 
        with u3d. Allowed values are: 
          * 0 ('globalmass'): concentration is transformed into a distribution
            of particles, by considering the cell volume, porosity and retardation factor.
            Distribution is obtained by transforming the total mass inside the
            cell into a number of particles estimated by means of particlesmass,
            distributed according to the cell aspect ratio, while using equispaced 
            and quasi-random particle distributions. Particles mass is unique for all 
            particles, recomputed in order to impose consistency between the total mass
            contained in the distribution and the estimated with particles.
          * 1 ('localmass'): similar to the previous alternative, particlesmass is employed 
            to estimate a number of particles per cell. The difference in this case, 
            is that once a number of particles is determined, particles mass is recomputed
            by establishing mass consistency at a cell level. This means that particles 
            from different cells may have different mass. Global mass consistency is 
            automatically satisfied.
    particlesmass  : float
        The scale of particles mass used for transforming the concentration data into 
        a distribution of particles. Determines resolution. 
    concentration : list, ndarray, int or float
        Interpreted as a resident initial concentration. It will be written with Util3d, 
        meaning that if the input is multidimensional, it should be consistent with 
        flow model dimensions. 
    particlesdist : int or str 
        The protocol for placing the particles inside each cell. Allowed values are: 
          * 0 ('equ','equispaced'): particles are placed following a uniform template
            determined from the estimated number of particles and the cell aspect 
            ratio. Particles are equispaced, and at half distance from the cell walls. 
          * 1 ('qra','quasirandom'): the equispaced template is perturbated based on
            a random number following a uniform distribution U[-1,1], proportional 
            to half the particle spacing.
          * 2 ('ran','random'): particles are placed randomly inside the cell. 
    drape : int 
        Determines the treatment of particles placed on initially dry cells. 
        Allowed values are: 
          * 0 : particles are unreleased. 
          * 1 : particles are vertically transfered to the next uppermost active cell. 
        notes:
          The drape option only applies for MODFLOW models solved with the standard formulation. 
          Models solved with Newton-Raphson are not affected by this parameter.
    speciesid : int 
        The species id to which this initial concentration is related. Will only 
        be written if particlesmassoption == 2. It should be given as the zero-based 
        index of the specie (the python-index of the specie in the list of species), 
        but is written to the package file with fortran format (+1, see the write_file method).
    stringid : str
        A string identifier. If None is given is automatically filled by 
        concatenating the package ftype and an integer id. 
    extension : str
        The file extension for the package. Defaults to ic. 
    """


    @staticmethod
    def _ftype():
        return 'IC'


    @multipackage
    def __init__(
        self,
        model,
        kind          = 0   ,
        particlesmass = 1.0 ,
        concentration = 0.0 ,
        particlesdist = 0   ,
        drape         = 0   ,
        speciesid     = 0   , 
        stringid      = None,
        extension     = 'ic',
    ):

        # Call parent constructor
        ftype = self._ftype()
        if ( model.multipackage[ftype]['count'] == 0 ): 
            super().__init__(
                model,
                extension,
                ftype,
                model.multipackage[ftype]['unitnumber']
            )
        else:
            # Pass the parent
            self._parent = model.multipackage[ftype]['instances'][0]._parent

        # Initial condition kind
        if ( not isinstance( kind, (int,str) ) ):
            raise TypeError(
                f"{self.__class__.__name__}:"
                f" Invalid type for initial condition kind."
                f" It can be specified as int or str, but {str(type(kind))}"
                f" was given." 
            )
        if ( isinstance( kind, int ) ):
            # Determines format for establishing mass consistency
            if (kind not in [0,1]): 
                raise ValueError(
                    f"{self.__class__.__name__}:"
                    f" Invalid initial condition kind parameter {str(kind)}."
                    f" Allowed values are 0 (globalmass) or 1 (localmass)."
                )
            self.kind = kind
        elif ( isinstance( kind, str ) ):
            try:
                self.kind = icKindOption[kind.lower()].value
            except:
                raise ValueError(
                    f"{self.__class__.__name__}:"
                    f" Invalid initial condition kind {str(kind)}."
                    f" The allowed values are globalmass or localmass." 
                )

        # The particles mass
        if ( not isinstance( particlesmass, (int,float) ) ):
            raise TypeError(
                f"{self.__class__.__name__}:"
                f" Invalid type for particlesmass. It should be int/float,"
                f" but {str(type(particlesmass))} was given."
            )
        if (particlesmass <= 0):
            raise ValueError(
                f"{self.__class__.__name__}:"
                f" Invalid particle mass for initial condition."
                f" It should be a positive value." 
            )
        self.particlesmass = particlesmass

        # Particles distribution format 
        if ( not isinstance( particlesdist, (int,str) ) ):
            raise TypeError(
                f"{self.__class__.__name__}:"
                f" Invalid type for particlesdist parameter."
                f" It can be specified as int or str, but {str(type(particlesdist))}"
                f" was given." 
            )
        if ( isinstance( particlesdist, int ) ):
            # Determines format for establishing mass consistency
            if (particlesdist not in [0,1,2]): 
                raise ValueError(
                    f"{self.__class__.__name__}:"
                    f" Invalid particlesdist parameter {str(particlesdist)}."
                    f" Allowed values are 0 (equispaced), 1 (quasirandom) "
                    f" or 2 (random)."
                )
            self.particlesdist = particlesdist
        elif ( isinstance( particlesdist, str ) ):
            try:
                self.particlesdist = particlesDistributionOption[particlesdist.lower()].value
            except:
                raise ValueError(
                    f"{self.__class__.__name__}:"
                    f" Invalid particlesdist parameter {str(particlesdist)}."
                    f" The allowed values are equ/equispaced, qra/quasirandom"
                    f" or ran/random."
                )

        # Drape option 
        if ( drape is not None ):
            if not isinstance( drape, int ): 
                raise TypeError(
                    f"{self.__class__.__name__}:"
                    f" Invalid type for drape. It should be integer,"
                    f" but {str(type(drape))} was given."
                )
            if (drape not in [0,1]):
                raise ValueError(
                    f"{self.__class__.__name__}:" 
                    f" Invalid value for drape. It should be"
                    f" 0 or 1."
                )
        else:
            drape = 0
        self.drape = drape

        # A species id, will only be written 
        # if particlesmassoption == 2
        if ( speciesid is not None ):
            if not isinstance( speciesid, int ): 
                raise TypeError(
                    f"{self.__class__.__name__}:"
                    f" Invalid type for speciesid. It should be integer,"
                    f" but {str(type(speciesid))} was given."
                )
            if (speciesid < 0):
                raise ValueError(
                    f"{self.__class__.__name__}:" 
                    f" Invalid value for speciesid. It should be"
                    f" a valid zero-based index of species."
                )
        self.speciesid = speciesid

        # String id 
        if (stringid is not None): 
            if ( not isinstance( stringid, str ) ): 
                raise TypeError(
                    f"{self.__class__.__name__}:"
                    f" Invalid type for stringid. It should be str, but"
                    f" {str(type(stringid))} was given."
                )
            self.stringid = stringid
        else:
            self.stringid = ftype+str(self.id)

        # and the concentration: needs id
        shape = model.shape
        if len(shape) == 3:
            shape3d = shape
        elif len(shape) == 2:
            shape3d = (shape[0], 1, shape[1])
        else:
            shape3d = (1, 1, shape[0])
        self.model = model
        self.shape3d = shape3d

        if (
            ( concentration is not None ) and 
            ( self.kind in [0,1] )
        ):
            # TypeError if any other unsupported type
            if ( not isinstance( concentration, (list,float,int,np.ndarray) ) ):
                raise TypeError(
                    f"{self.__class__.__name__}:"
                    f" Invalid type {str(type(concentration))}"
                    f" for concentration. It should be list, np.ndarray or a single numeric value."
                )
            # If list, to np.array
            if ( isinstance( concentration, list ) ): 
                concentration = np.array( concentration )
            # Verify a valid dtype
            if ( isinstance( concentration, np.ndarray ) ): 
                if ( 
                    (concentration.dtype not in [np.int32, np.int64, np.float32, np.float64])
                ):
                    raise TypeError(
                        f"{self.__class__.__name__}:"
                        f" Invalid dtype {str(concentration.dtype)}" 
                        f" for concentration. It should be int or float."
                    )
            try:
                self.concentration = Util3d(
                    model,
                    shape3d,
                    np.float32,
                    concentration,
                    name=self.stringid,
                    locat=self._parent.multipackage[ftype]['unitnumber'], 
                )
            except:
                raise Exception(
                    f"{self.__class__.__name__}:"
                    f" Error while initializing distributed variable concentration."
                    f" Is the input shape consistent with flow model dimensions ?"
                )

        elif ( (self.kind in [0,1]) and (concentration is None) ):
            raise ValueError(
                f"{self.__class__.__name__}:"
                f" Invalid initial concentration. None was given."
            )
    

        # Done
        return


    def write_file(self, check=False):
        """
        Write the package file
        Parameters
        ----------
        check : boolean
            Check package data for common errors. (default False)
        Returns
        -------
        None
        """

        # Shortcut to multipackage dict
        mpkg = self.parent.multipackage[self._ftype()]
        
        # Write
        with open( mpkg['instances'][0].fn_path, 'w' ) as f:

            # Write how many instances...
            f.write(f"{mpkg['count']}\n")

            # ...and loop over them
            for ins in mpkg['instances']:

                # Write string id 
                f.write(f"{ins.stringid}\n")

                # Kind/format of initial condition, particles distribution and drape
                f.write(f"{ins.kind}   {ins.particlesdist}   {ins.drape}\n")

                # 0,1: resident concentration array  
                if ( ins.kind in [0,1] ):

                    # Give particles mass 
                    f.write(f"{ins.particlesmass:.10f}\n")

                    # If solute ids are being specified, write the id
                    if( 
                        ( self._parent.particlesmassoption == 2 ) 
                    ):
                        if ( 
                            ( ins.speciesid is None ) or
                            ( ins.speciesid < 0 ) 
                        ):
                            raise ValueError(
                                f"{self.__class__.__name__}:"
                                f" Invalid species id for initial condition."
                                f" It should be a valid index in the list of species."
                            )
                        f.write(f"{ins.speciesid+1}\n")

                    # And the concentration distribution
                    f.write(ins.concentration.get_file_entry())

        # Done
        return



'''
Configuration of MODPATH-RW impermeable cells
'''

# python
import numpy as np

# flopy
from flopy.pakbase import Package
from flopy.utils import Util3d


class ModpathRWImp( Package ):
    """
    MODPATH-RW Impermeable Cells Class.

    Parameters
    ----------
    model : model object
        The model object (of type :class: ModpathRW) to which
        this package will be added.
    defaultboundary : int
        The default behavior of particles arriving to cell faces
        without connections. Allowed values are: 
         * 0: boundaries are considered open and particles are removed
         * 1: boundaries are considered closed and particles rebound
    impformat : int
        The format in which impermeable cells are defined. Allowed values are: 
         * 0: Follow flow-model inactive cells 
         * 1: Follow time-dependent flow-model inactive cells (include dry cells)
         * 2: Read icbound parameter and write with u3d 
    icbound : list, np.ndarray [int]
        Specification for impermeable cells. Particles rebound 
        in those cells with value 1. Read with u3d. 
    """
   
    
    @staticmethod
    def _ftype():
        return 'IMP'


    def __init__(
            self,
            model,
            impformat       = 0,
            defaultboundary = 0,
            icbound         = 0, # for u3d
            extension       = 'imp',
        ):
  
        # Call super constructor
        ftype      = self._ftype()
        unitnumber = model.next_unit()
        super().__init__(model, extension, ftype, unitnumber)
      
        # Default boundary 
        if ( defaultboundary not in [0,1] ): 
            raise ValueError(
                f"{self.__class__.__name__}:"
                f" Invalid defaultboundary {str(defaultboundary)}"
                f". Allowed values are 0 (open boundaries) "
                f"or 1 (impermeable boundaries)."
            )
        self.defaultboundary = defaultboundary

        # Impermeable cells format 
        if ( impformat not in [0,1,2] ): 
            raise ValueError(
                f"{self.__class__.__name__}:"
                f" Invalid impformat {str(impformat)}."
                f" Allowed values are 0 (follow ibound), "
                f"1 (follow iboundts) or 2 (given in icbound param)."
            )
        self.impformat = impformat

        # The case where icbound is needed 
        if self.impformat == 2:

            if icbound is None:
                raise ValueError(
                    f"{self.__class__.__name__}:"
                    f" Impermeable cells format was " 
                    f"specified as 2, but no icbound was given."
                )

            shape = model.shape
            if len(shape) == 3:
                shape3d = shape
            elif len(shape) == 2:
                shape3d = (shape[0], 1, shape[1])
            else:
                shape3d = (1, 1, shape[0])
            self.model = model
            self.shape3d = shape3d

            if not isinstance( icbound, (int,list,np.ndarray) ):
                raise TypeError(
                    f"{self.__class__.__name__}:"
                    f" Invalid type for icbound. It should be list, or np.ndarray. "
                    f"{str(type(icbound))} was given."
                )
            if isinstance( icbound, int ): 
                icbound = [ icbound ]
            icbound  = np.array(icbound)
            uicbound = np.unique(icbound)
            for uc in uicbound: 
                if uc not in [0,1]:
                    raise ValueError( 
                        f"{self.__class__.__name__}:"
                        f" Invalid values for icbound specification. "
                        f"It should only contain values 0 or 1. "
                        f"The value {str(uc)} was found."
                    )

            try:
                self.icbound= Util3d(
                    model,
                    shape3d,
                    np.int32,
                    icbound,
                    name="ICBOUND",
                    locat=self.unit_number[0],
                )
            except:
                raise Exception(
                    f"{self.__class__.__name__}:"
                    f" Error while initializing distributed variable icbound. "
                    f"Is the input shape consistent with flow model dimensions ?"
                )

        # Add package 
        self.parent.add_package(self)

        # Done 
        return


    def write_file(self, check=False):
        """
        Write the package file

        Parameters
        ----------
        check : boolean
            Check package data for common errors. (default False)
        Returns
        -------
        None
        """

        # Open file for writing
        with open(self.fn_path, 'w') as f:

            # Write the default boundary 
            f.write(f"{self.defaultboundary}\n")  

            # Write the impermeable cells format 
            f.write(f"{self.impformat}\n")

            # For impformat 2, write icbound
            if ( self.impformat == 2):
                f.write(self.icbound.get_file_entry())

        # Done
        return



'''
Implement classes and methods for configuring MODPATH-RW sims with flopy
'''

from .mprw    import ModpathRW
from .mprwsim import ModpathRWSim 
from .mprwbas import ModpathRWBas 
from .mprwparticlegroup import (
        ParticleGroup, 
        ParticleGroupLRCTemplate, 
        ParticleGroupNodeTemplate
    )


from .mprwobs   import ModpathRWObs
from .mprwdsp   import ModpathRWDsp
from .mprwopts  import ModpathRWOpts
from .mprwspc   import ModpathRWSpc
from .mprwic    import ModpathRWIc
from .mprwimp   import ModpathRWImp
from .mprwsrc   import ModpathRWSrc
from .mprwgpkde import ModpathRWGpkde



'''
Configuration of MODPATH-RW gpkde reconstruction
'''

# python
import numpy as np
from enum import Enum

# flopy
from flopy.utils import Util2d
from flopy.pakbase import Package
from flopy.discretization import StructuredGrid, VertexGrid, UnstructuredGrid


class slicedDimensionOption(Enum): 
    '''
    Enumerate possible values for sliced dimension 
    '''
    x = 0
    y = 1
    z = 2


class ModpathRWGpkde( Package ):
    """
    MODPATH-RW GPKDE Package Class.

    Parameters
    ----------
    model : model object
        The model object (of type :class: ModpathRW) to which
        this package will be added.
    outputfilename: str
        Filename that will be used for writing reconstruction output.
    outputcolformat : int
        Determines the columns structure of the output file. Allowed values are: 
          * 0: bin ids and density data
          * 1: bin ids, cell coordinates and density data
          * 2: cell coordinates and density data
    outputfileformat : int
        Determines the format of the output file. Allowed values are: 
          * 0: text plain file
          * 1: binary unformatted file
    domainorigin : list [ox,oy,oz]
        List with domain origin. If not given will be set as default to 
        [ xoffset, yoffset, min(botm) ]
    domainsize : list [sx,sy,sz]
        List with domain sizes. These are used for computing reconstruction 
        grid dimensions using given binsize.
    binsize : list [sx,sy,sz]
        List with cell sizes employed for reconstruction process. If the size on 
        a given dimension is zero, then the histogram count will not consider said 
        dimension.
    gridallocformat : int
        Determines the allocation in memory of the reconstruction grid. Allowed values are: 
          * 0: allocate following domain size 
          * 1: allocate according to the particle distribution
    gridborderfraction : float
        While allocating according to the particle locations, border fraction determines 
        an additional allocated space surrounding the particle cloud, extends the 
        bounding box allowing the boundaries to receive information from the interior
        histogram cells. For a given dimension, the distance considered as buffer
        is equal to gridborderfraction*extentdimension and is distributed as half
        for each side. Should be between [0,1].
    slicedreconstruction : bool
        Performs reconstruction for each slice of a given dimension. Recommended 
        for quasi two-dimensional problems, for example, when the horizontal extent
        dominates over the vertical. 
    sliceddimension: int or str
        The dimension to be sliced. Allowed values are:
          0 or 'x': reconstruction for each slice of dimension x 
          1 or 'y': reconstruction for each slice of dimension y 
          2 or 'z': reconstruction for each slice of dimension z
    noptloops : int
        The number of bandwidth optimization loops. It should be a positive integer.
    skiperror : bool
        Flag to indicate whether the error checks for the optimization of bandwidth
        selection should be skipped or not. If skipped, optimization is performed 
        until noptloops.
    convergence : float
        The limit relative change of the variables monitored during bandwidth optimization 
        to determine the convergence of reconstruction. It should be positive.
    kerneldatabase : bool
        Flag to indicate whether reconstruction should be performed with a database of 
        kernels (True) or by computing kernels in real time (False). In case a kernel 
        database is allocated, then the class will employ the parameters 
        minhd: deltahd: maxhd to determine a range of discrete kernel sizes for
        allocation. This range is relative to the bin size, meaning, 
        parameters *hd indicate the ratio smoothing(h)/binsize(d) for each model dimension.
    isotropickernels : bool
        Flag to indicate that kernels are considered as isotropic. This consideration 
        is done in terms of the kernel smoothing (h), meaning that if the bin sizes 
        are non-isotropic, then the program will force the same kernel smoothing taking 
        into account the different cell sizes.
    anisotropickernelsic : int
        Indicate the isotropy status for kernels employed in the reconstruction of the 
        initial condition. Allowed values are:
          * 0: kernels are isotropic.
          * 1: kernels are anisotropic.
          * 2: kernels follow the variable isotropickernels.
    boundkernelsize : int
        Determines the protocol for bounding the kernel size. Allowed values are: 
          * 0: bound size by domain constraints. 
          * 1: bound using minhd and maxhd
          * 2: unbounded
    minhd : float
        Minimum ratio smoothing(h)/binsize(l). It is written to the package file 
        if kerneldatabase=True or boundkernelsize = 1.
    deltahd : float
        Step of ration smoothing(h)/binsize(l) used for defining a set of discrete
        kernel sizes for the kernel database. Only written to the package file 
        if kerneldatabase=True.
    maxhd : float
        Maximum ration smoothing(h)/binsize(l). It is written to the package file if
        kerneldatabase=True or boundkernelsize = 1.
    initialsmoothingformat : int
        Determines the protocol for selection of the initial kernel size from where to
        begin the optimization for bandwidth selection. Allowed values are: 
          * 0: selects initial bandwidth from the expression of Silverman (1986) 
               for Gaussian distributions. It is based on the standard deviation of
               the particle coordinates, and the number of particles. 
          * 1: initial bandwidth as a factor multiplying the characteristic cell size.
               It will use the parameter binsizefactor to perform the scaling.
    binsizefactor : float
        Scaling factor employed for determining the initial kernel size from where to start
        the bandwidth optimization. Amplifies the characteristic cell size.
    asconcentration: bool
        Flag to indicate whether the density reconstruction should be returned as 
        resident concentration (dissolved). Because the reconstruction grid is different
        than the flowmodel grid, said transformation can be easily achieved only in case
        both porosities and delaying factors are spatially uniform. Otherwise, density is 
        returned as total mass concentration, including both sorbed and aqueous phases.
    effectiveweightformat : int
        Determines the protocol for handling the reconstruction of particle distributions 
        with different weights. Allowed values are: 
          * 0: computes a unique effective number of particles and determines a 
               domain-effective particle weight (Kish, 1965,1992), which is used to
               transform the mass histogram into an equivalent particle distribution 
               used for the bandwidth selection.
          * 1: similar to the previous alternative, but the employed characteristic 
               particle weight is the average over all particles. 
          * 2: Bandwidth selection is performed taking into account only the particle 
               coordinates and a final density estimation is performed over the mass 
               histogram with the obtain distribution of kernel sizes.
          * 3: An effective number of particles is computed for each histogram cell and 
               is employed to transform the mass distribution into an effective particle 
               distribution used for the bandwidth optimization. A final reconstruction
               stage is performed considering the obtained distribution of kernel sizes.
    timepointdata : list or tuple
        Follows a similar logic to timepointdata at the ModpathRWSim/Modpath7Sim class.
        Is a list or tuple with 2 items. If the second item is a float then 
        thetimepointdata corresponds to timepointoption 1 and the first entry is the
        number of time points (timepointcount) and the second the interval.
        If the second item is a list, tuple, or np.ndarray, timepointoption is 2 and
        reconstruction is performed for the times of the given list. If timepointdata is None,
        timepointoption is 0 and reconstruction follows the timeseries points. 
    skipinitialcondition: bool
        Flag to indicate whether reconstruction should be performed for the 
        initial distribution of particles.
    extension : str, optional
        File extension (default is 'gpkde').
    """

    
    @staticmethod
    def _ftype():
        return 'GPKDE'


    def __init__(
        self,
        model,
        outputfilename         = None,
        outputcolformat        = 0, 
        outputfileformat       = 0, 
        domainorigin           = None, 
        domainsize             = None, 
        binsize                = None,
        gridallocformat        = 0, 
        gridborderfraction     = 0.05,
        slicedreconstruction   = None, 
        sliceddimension        = None, 
        noptloops              = 10,
        skiperror              = False, 
        convergence            = 0.02,
        kerneldatabase         = False, 
        isotropickernels       = False,
        boundkernelsize        = 0, 
        anisotropickernelsic   = 0, 
        minhd                  = 1.0 ,
        maxhd                  = 0.1 ,
        deltahd                = 10.0,
        initialsmoothingformat = 0, 
        binsizefactor          = 1.0, 
        asconcentration        = True,
        effectiveweightformat  = 0,
        timepointdata          = None,
        skipinitialcondition   = False,
        extension              = 'gpkde',
    ):
        
        # initialize
        ftype = self._ftype()
        unitnumber = model.next_unit()
        super().__init__(model, extension, ftype, unitnumber)
        
        # outputfilename
        if outputfilename is not None:
            if ( 
                not isinstance( outputfilename, str )  
            ):
                raise TypeError(
                    f"{self.__class__.__name__}:" 
                    f" The type of the output file name should be str."
                    f" {str(type(outputfilename))} was given."
                )
            self.outputfilename = outputfilename
        else:
            self.outputfilename = f"{model.name}.gpkde.out"

        # outputcolformat
        if outputcolformat not in [0,1,2]: 
            raise ValueError(
                f"{self.__class__.__name__}:"
                f" Invalid value for outputcolformat. Allowed values are"
                f" 0 (bin ids and density), 1 (bin ids, coordinates and density)"
                f" and 2 (coordinates and density)."
                f" {str(outputcolformat)} was given."
            )
        self.outputcolformat = outputcolformat

        # outfileformat
        if outputfileformat not in [0,1]: 
            raise ValueError(
                f"{self.__class__.__name__}:"
                f" Invalid value for outputfileformat. Allowed values are"
                f" 0 (text-plain) or 1 (binary)."
                f" {str(outputfileformat)}  was given."
            )
        self.outputfileformat = outputfileformat


        # Get the modelgrid for convenience
        modelgrid = self.parent.flowmodel.modelgrid 

        # domainsize
        if (
            ( domainsize is None ) and 
            isinstance( modelgrid, (StructuredGrid,VertexGrid) )
        ):

            domainsize = np.zeros(shape=(3,)).astype(np.float32)
            xmin,xmax,ymin,ymax = modelgrid.extent

            if ( isinstance( modelgrid, StructuredGrid ) ):
                # Assign domainsizes
                domainsize[0] = abs( xmax - xmin )
                domainsize[1] = abs( ymax - ymin )
                zmin = np.min( modelgrid.botm )
                zmax = np.max( modelgrid.top  ) 
                domainsize[2] = abs( zmax - zmin )
            elif ( isinstance( modelgrid, VertexGrid ) ):
                # Is the same than for structured 
                domainsize[0] = abs( xmax - xmin )
                domainsize[1] = abs( ymax - ymin )
                zmin = np.min( modelgrid.botm )
                zmax = np.max( modelgrid.top  ) 
                domainsize[2] = abs( zmax - zmin )
        elif (
            ( domainsize is None ) and 
            isinstance( modelgrid, (UnstructuredGrid) )
        ):
            # In the meantime, force explicit specification
            if ( self.parent.flowmodel.modelversion == 'mf6' ):
                raise ValueError(
                    f"{self.__class__.__name__}:"
                    f" For unstructured grids an explicit specification"
                    f" of domainsize is needed, but None was given."
                    f" Note: Modpath7 and ModpathRW do not currently support"
                    f" unstructured grids for mf6 flow models."
                )
            else: 
                raise ValueError(
                    f"{self.__class__.__name__}:"
                    f" For unstructured grids an explicit specification"
                    f" of domainsize is needed, but None was given."
                )

        if ( not isinstance( domainsize, (list,np.ndarray) ) ):
            raise TypeError(
                f"{self.__class__.__name__}:" 
                f" The type of the domainsize parameter"
                f" should be list or np.ndarray."
                f" {str(type(domainsize))} was given."
            )
        domainsize = np.array(domainsize).astype(np.float32)
        if ( len(domainsize.shape) != 1 ):
            raise ValueError(
                f"{self.__class__.__name__}:" 
                f" Invalid specification of domain size."
                f" It should be a one dimensional list with 3 elements."
            )
        if ( domainsize.shape[0] != 3 ):
            raise ValueError(
                f"{self.__class__.__name__}:"
                f" Invalid specification of domain size."
                f" It should be a one dimensional list with 3 elements."
            )
        if ( np.any( domainsize < 0 ) ): 
            raise ValueError(
                f"{self.__class__.__name__}:"
                f" Bin sizes should be greater than zero."
                f" {str(self.domainsize)} was given."
            )
        if ( np.all( domainsize == 0 ) ): 
            raise ValueError(
                f"{self.__class__.__name__}:"
                f" All domain sizes are zero."
                f" At least one positive value should be given."
            )
        self.domainsize = domainsize


        # domainorigin
        if (
            ( domainorigin is None ) and 
            isinstance( modelgrid, (StructuredGrid,VertexGrid) )
        ):

            domainorigin = np.zeros(shape=(3,)).astype(np.float32)

            if ( isinstance( modelgrid, StructuredGrid ) ):
                # Fills domain origin with values 
                # determined from the modelgrid definition
                domainorigin[0] = modelgrid.xoffset
                domainorigin[1] = modelgrid.yoffset
                domainorigin[2] = np.min(modelgrid.botm)
            elif ( isinstance( modelgrid, VertexGrid ) ):
                # The same as structured 
                domainorigin[0] = modelgrid.xoffset
                domainorigin[1] = modelgrid.yoffset
                domainorigin[2] = np.min(modelgrid.botm)
        elif (
            ( domainorigin is None ) and 
            isinstance( modelgrid, (UnstructuredGrid) )
        ):
            # In the meantime, force explicit specification
            if ( self.parent.flowmodel.modelversion == 'mf6' ):
                raise ValueError(
                    f"{self.__class__.__name__}:"
                    f" For unstructured grids an explicit specification"
                    f" of domainorigin is needed, but None was given."
                    f" Note: Modpath7 and ModpathRW do not currently support"
                    f" unstructured grids for mf6 flow models."
                )
            else: 
                raise ValueError(
                    f"{self.__class__.__name__}:"
                    f" For unstructured grids an explicit specification"
                    f" of domainorigin is needed, but None was given."
                )

        if ( not isinstance( domainorigin, (list,np.ndarray) ) ):
            raise TypeError(
                f"{self.__class__.__name__}:" 
                f" The type of the domainorigin parameter"
                f" should be list or np.ndarray."
                f" {str(type(domainorigin))} was given."
            )
        domainorigin = np.array(domainorigin).astype(np.float32)
        if ( len(domainorigin.shape) != 1 ):
            raise ValueError(
                f"{self.__class__.__name__}:"
                f" Invalid specification of domain origin."
                f" It should be a one dimensional list with 3 elements."
            )
        if ( domainorigin.shape[0] != 3 ): 
            raise ValueError(
                f"{self.__class__.__name__}:"
                f" Invalid specification of domain origin."
                f" It should be a one dimensional list with 3 elements."
            )
        self.domainorigin = domainorigin


        # binsize
        if (
            ( binsize is None ) and 
            isinstance( modelgrid, (StructuredGrid,VertexGrid) )
        ): 
            # For a structured grid, infer bin sizes
            # from delr,delc,delz
            binsize = np.zeros(shape=(3,)).astype(np.float32)

            if ( isinstance( modelgrid, StructuredGrid ) ):
                # x
                if modelgrid.is_regular_x:
                    # If the grid is regular, then get the value
                    binsize[0] = modelgrid.delr.flat[0]
                else:
                    # If not, it could be an average
                    binsize[0] = np.mean(modelgrid.delr.flat)
                # y
                if modelgrid.is_regular_y:
                    binsize[1] = modelgrid.delc.flat[0]
                else:
                    binsize[1] = np.mean(modelgrid.delc.flat)
                # z
                if modelgrid.is_regular_z:
                    binsize[2] = modelgrid.delz.flat[0]
                else:
                    binsize[2] = np.mean(modelgrid.delz.flat)
            elif ( isinstance( modelgrid, VertexGrid ) ):
                # Infer a bin size based on xcellcenters 
                # and ycellcenters
                # x
                # Check if unique difference
                xunique = np.unique(modelgrid.xcellcenters)
                if len(xunique) > 1:
                    xdiff = np.diff(xunique)
                    if ( len(np.unique(xdiff))==1 ):
                        # Take as bin size if unique
                        binsize[0] = np.unique(xdiff).item()
                    else:
                        # otherwise an average
                        binsize[0] = np.mean(np.unique(xdiff))
                else:
                    # if one element, assign domainsize
                    binsize[0] = domainsize[0]
                # y
                yunique = np.unique(modelgrid.ycellcenters)
                if len(yunique) > 1:
                    ydiff = np.diff(yunique)
                    if ( len(np.unique(ydiff))==1 ):
                        binsize[1] = np.unique(ydiff).item()
                    else:
                        binsize[1] = np.mean(np.unique(ydiff))
                else:
                    # if one, assign domainsize
                    binsize[1] = domainsize[1]
                # z
                zunique = np.unique(modelgrid.zcellcenters)
                if len(zunique) > 1:
                    zdiff = np.diff(zunique)
                    if ( len(np.unique(zdiff))==1 ):
                        binsize[2] = np.unique(zdiff).item()
                    else:
                        binsize[2] = np.mean(np.unique(zdiff))
                else:
                    # if one, assign domainsize
                    binsize[2] = domainsize[2]
        elif (
            ( binsize is None ) and 
            isinstance( modelgrid, (UnstructuredGrid) )
        ):
            # In the meantime, force explicit specification
            if ( self.parent.flowmodel.modelversion == 'mf6' ):
                raise ValueError(
                    f"{self.__class__.__name__}:"
                    f" For unstructured grids an explicit specification"
                    f" of binsize is needed, but None was given."
                    f" Note: Modpath7 and ModpathRW do not currently support"
                    f" unstructured grids for mf6 flow models."
                )
            else: 
                raise ValueError(
                    f"{self.__class__.__name__}:"
                    f" For unstructured grids an explicit specification"
                    f" of binsize is needed, but None was given."
                )

        if ( not isinstance( binsize, (list,np.ndarray) ) ):
            raise TypeError(
                f"{self.__class__.__name__}:"
                f" The type of the binsize parameter should be list or np.ndarray." 
                f" {str(type(binsize))}  was given."
            )
        binsize = np.array(binsize).astype(np.float32)
        if ( len(binsize.shape) != 1 ):
            raise ValueError(
                f"{self.__class__.__name__}:"
                f" Invalid specification of bin size."
                f" It should be a one dimensional list with 3 elements."
            )
        if ( binsize.shape[0] != 3 ): 
            raise ValueError(
                f"{self.__class__.__name__}:"
                f" Invalid specification of bin size."
                f" It should be a one dimensional list with 3 elements."
            )
        if ( np.any( binsize < 0 ) ): 
            raise ValueError(
                f"{self.__class__.__name__}:"
                f" Bin sizes should be greater than zero."
                f" {str(self.binsize)} was given."
            )
        if ( np.all( binsize == 0 ) ): 
            raise ValueError(
                f"{self.__class__.__name__}:"
                f" All bin sizes are zero."
                f" At least one positive value should be given."
            )
        self.binsize = binsize

        # gridallocformat
        if (gridallocformat not in [0,1]):
            raise ValueError(
                f"{self.__class__.__name__}:"
                f" Invalid value for gridallocformat. Allowed values are"
                f" 0 (allocate the grid as the domain) or 1 (allocate adapting to the particles)."
                f" {str(gridallocformat)} was given."
            )
        self.gridallocformat = gridallocformat

        # gridborderfraction
        if ( self.gridallocformat ==  1 ):
            # validate only if gridallocformat = 1
            if ( (not 0<gridborderfraction<=1) ):
                raise ValueError(
                    f"{self.__class__.__name__}:"
                    f" Invalid value for gridborderfraction. Should be between 0 and 1."
                    f" 0 (allocate the grid as the domain) or 1 (allocate adapting to the particles)."
                    f" {str(gridborderfraction)} was given."
                )
            self.gridborderfraction = gridborderfraction

        # sliced reconstruction
        if ( slicedreconstruction is None ): 
            # some inference for 3d problems
            if ( np.all( binsize >  0 ) ):
                # rough estimate of the number of bins
                nbins = domainsize/binsize
                # skip inference if any nbins is one, 
                # gpkde will recognize this case.
                if np.any( nbins == 1 ): 
                    self.slicedreconstruction = False
                    self.sliceddimension = 0
                else:
                    # Get the minimum number of bins and estimate 
                    # anisotropy
                    idmin, = np.where( nbins == np.min( nbins ) )
                    idmin  = idmin.item()
                    # none of these seems to be that general
                    ani13  = nbins[(idmin+1)%3]/nbins[idmin]
                    ani23  = nbins[(idmin+2)%3]/nbins[idmin]
                    #ani13  = domainsize[(idmin+1)%3]/domainsize[idmin]
                    #ani23  = domainsize[(idmin+2)%3]/domainsize[idmin]

                    # Define an arbitrary threshold of 10
                    if ( ( ani13 > 10 ) and ( ani23 > 10 ) ) :
                        self.slicedreconstruction = True
                        self.sliceddimension = idmin
                    else:
                        # the inference might be improved, but in the meantime
                        # continue as false.
                        self.slicedreconstruction = False
                        self.sliceddimension = 0
            else:
                self.slicedreconstruction = False
                self.sliceddimension = 0
        else:
            if ( not isinstance( slicedreconstruction, bool ) ) : 
                raise TypeError(
                    f"{self.__class__.__name__}:"
                    f" Invalid type for the slicedreconstruction parameter. It should be bool, but"
                    f" {str(type(slicedreconstruction))} was given."
                )
            self.slicedreconstruction = slicedreconstruction

            # process sliceddimension
            if ( self.slicedreconstruction ): 
                if ( sliceddimension is None ):
                    # Do the inference as before
                    # some inference for 3d problems
                    if ( np.all( binsize >  0 ) ):
                        # rough estimate of the number of bins
                        nbins = domainsize/binsize
                        # skip inference if any nbins is one, 
                        # gpkde will recognize this case.
                        if np.any( nbins == 1 ):
                            # force the user to provide a value
                            raise ValueError(
                                f"{self.__class__.__name__}:"
                                f" Invalid value for sliceddimension. None was provided and"
                                f" inference could not be done."
                            )

                        else:
                            # Get the minimum number of bins and estimate 
                            # anisotropy
                            idmin, = np.where( nbins == np.min( nbins ) )
                            idmin  = idmin.item()
                            # none of these seems to be that general
                            ani13  = nbins[(idmin+1)%3]/nbins[idmin]
                            ani23  = nbins[(idmin+2)%3]/nbins[idmin]
                            #ani13  = domainsize[(idmin+1)%3]/domainsize[idmin]
                            #ani23  = domainsize[(idmin+2)%3]/domainsize[idmin]

                            # Define an arbitrary threshold of 10
                            if ( ( ani13 > 10 ) and ( ani23 > 10 ) ) :
                                self.slicedreconstruction = True
                                self.sliceddimension = idmin
                            else:
                                # the inference might be improved, but in the meantime
                                # force the user to provide a value
                                raise ValueError(
                                    f"{self.__class__.__name__}:"
                                    f" Invalid value for sliceddimension. None was provided and"
                                    f" inference could not be done."
                                )

                elif ( not isinstance( sliceddimension, (int,str) ) ):
                    raise TypeError(
                        f"{self.__class__.__name__}:"
                        f" Invalid type for the sliceddimension parameter. It should be int or str, but"
                        f" {str(type(sliceddimension))} was given."
                    )
                else:
                    if isinstance( sliceddimension, int ) :
                        if ( sliceddimension not in [0,1,2] ):
                            raise ValueError(
                                f"{self.__class__.__name__}:"
                                f" Invalid sliceddimension {str(sliceddimension)}."
                                f" The allowed values are 0 (x), 1 (y) or 2 (z)."
                            )
                        self.sliceddimension = sliceddimension + 1 
                    else:
                        try:
                            self.sliceddimension = slicedDimensionOption[sliceddimension.lower()].value + 1
                        except:
                            raise ValueError(
                                f"{self.__class__.__name__}:"
                                f" Invalid sliceddimension {str(sliceddimension)}."
                                f" The allowed values are x, y or z."
                            )

        # noptloops
        if ( not isinstance( noptloops, int) ): 
            raise TypeError(
                f"{self.__class__.__name__}:"
                f" Invalid type for the number of optimization loops. It should be int, but"
                f" {str(type(noptloops))} was given."
            )
        if ( noptloops < 0 ): 
            raise ValueError(
                f"{self.__class__.__name__}:"
                f" Invalid value for the number of optimization loops. It should be a positive integer, but"
                f" {str(noptloops)} was given."
            )
        self.noptloops = noptloops

        # skiperror
        if ( not isinstance( skiperror, bool ) ): 
            raise TypeError(
                f"{self.__class__.__name__}:"
                f" Invalid type for the skiperror parameter. It should be a boolean, but"
                f" {str(type(skiperror))} was given."
            )
        self.skiperror = skiperror

        # convergence
        if ( not isinstance( convergence, (int,float) ) ): 
            raise TypeError(
                f"{self.__class__.__name__}:"
                f" Invalid type for the number of optimization loops. It should be float, but"
                f" {str(type(convergence))} was given."
            )
        if ( convergence < 0 ): 
            raise ValueError(
                f"{self.__class__.__name__}:"
                f" Invalid value for the convergence parameter. It should positive, but"
                f" {str(convergence)} was given."
            )
        self.convergence = convergence

        # kerneldatabase
        if ( not isinstance( kerneldatabase, bool ) ): 
            raise TypeError(
                f"{self.__class__.__name__}:"
                f" Invalid type for the kerneldatabase parameter. It should be a boolean, but"
                f" {str(type(kerneldatabase))} was given."
            )
        self.kerneldatabase = kerneldatabase

        # isotropickernels
        if ( not isinstance( isotropickernels, bool ) ): 
            raise TypeError(
                f"{self.__class__.__name__}:" 
                f" Invalid type for the isotropickernels parameter. It should be a boolean, but"
                f" {str(type(isotropickernels))} was given."
            )
        self.isotropickernels = isotropickernels

        # anisotropickernelsic
        if ( not isinstance( anisotropickernelsic, int ) ): 
            raise TypeError(
                f"{self.__class__.__name__}:" 
                f" Invalid type for the anisotropickernelsic parameter. It should be int, but"
                f" {str(type(anisotropickernelsic))} was given."
            )
        if ( anisotropickernelsic not in [0,1,2] ): 
            raise ValueError(
                f"{self.__class__.__name__}:" 
                f" Invalid value for anisotropickernelsic . Allowed values are 0 (isotropic kernels for ic),"
                f" 1 (anisotropic kernels for ic) or 2 (follow isotropickernels)."
                f" {str(anisotropickernelsic)} was given."
            )
        self.anisotropickernelsic = anisotropickernelsic

        # boundkernelsize
        if ( boundkernelsize not in [0,1,2] ): 
            raise ValueError(
                f"{self.__class__.__name__}:" 
                f" Invalid value for boundkernelsize. Allowed values are 0 (bound by domain constraints),"
                f" 1 (bound by minhd,maxhd) or 2 (unbounded)."
                f" {str(boundkernelsize)} was given."
            )
        self.boundkernelsize = boundkernelsize

        # minhd
        # Verify only if required 
        if ( ( self.kerneldatabase ) or ( self.boundkernelsize==1 ) ):
            if ( not isinstance( minhd, (float,int) ) ):
                raise TypeError(
                    f"{self.__class__.__name__}:" 
                    f" Invalid type for minhd. It should be float or int, but"
                    f" {str(type(minhd))} was given."
                )
            if ( minhd <= 0 ):
                raise ValueError(
                    f"{self.__class__.__name__}:" 
                    f" Invalid value for minhd. It should be a positive value, but"
                    f" {str(type(minhd))} was given."
                )
            self.minhd = minhd

        # deltahd
        # Verify only if required 
        if ( ( self.kerneldatabase ) ):
            if ( not isinstance( deltahd, (float,int) ) ):
                raise TypeError(
                    f"{self.__class__.__name__}:"
                    f" Invalid type for deltahd. It should be float or int, but"
                    f" {str(type(deltahd))} was given."
                )
            if ( deltahd <= 0 ):
                raise ValueError(
                    f"{self.__class__.__name__}:" 
                    f" Invalid value for deltahd. It should be a positive value, but"
                    f" {str(type(deltahd))} was given."
                )
            self.deltahd = deltahd

        # maxhd
        # Verify only if required 
        if ( ( self.kerneldatabase ) or ( self.boundkernelsize==1 ) ):
            if ( not isinstance( maxhd, (float,int) ) ):
                raise TypeError(
                    f"{self.__class__.__name__}:" 
                    f" Invalid type for maxhd. It should be float or int, but"
                    f" {str(type(maxhd))} was given."
                )
            if ( maxhd <= 0 ):
                raise ValueError(
                    f"{self.__class__.__name__}:" 
                    f" Invalid value for maxhd. It should be a positive value, but"
                    f" {str(type(maxhd))} was given."
                )
            self.maxhd = maxhd

        # Minor consistency checks
        if ( ( self.kerneldatabase ) or ( self.boundkernelsize==1 ) ):
            if ( not( self.minhd <= self.maxhd ) ): 
                raise ValueError(
                    f"{self.__class__.__name__}:" 
                    f" Inconsistent values for minhd and maxhd."
                    f" Maxhd should be greater than minhd."
                )
            if ( self.kerneldatabase ):
                if ( ( self.deltahd > self.maxhd ) ): 
                    raise ValueError(
                        f"{self.__class__.__name__}:" 
                        f" Inconsistent values for deltahd and maxhd."
                        f" The given step value is higher than the maxhd."
                    )

        # initialsmoothingformat
        if (initialsmoothingformat not in [0,1]):
            raise ValueError(
                f"{self.__class__.__name__}:" 
                f" Invalid value for initialsmoothingformat. Allowed values are " 
                f" 0 (automatic initial selection) or 1 (initial kernel as a factor amplifying the bin size)."
                f" {str(initialsmoothingformat)} was given."
            )
        self.initialsmoothingformat = initialsmoothingformat

        # binsizefactor
        # Validate only if necessary
        if ( self.initialsmoothingformat == 1 ): 
            if ( not isinstance( binsizefactor, ( int, float ) ) ):
                raise TypeError(
                    f"{self.__class__.__name__}:" 
                    f" Invalid type for binsizefactor. It should be float or int, but"
                    f" {str(type(binsizefactor))} was given."
                )
            if ( binsizefactor <= 0 ):
                raise ValueError(
                    f"{self.__class__.__name__}:" 
                    f" Invalid value for binsizefactor. It should be a positive value, but"
                    f" {str(binsizefactor)} was given."
                )
            self.binsizefactor = binsizefactor

        # asconcentration
        if ( not isinstance( asconcentration, bool ) ): 
            raise TypeError(
                f"{self.__class__.__name__}:" 
                f" Invalid type for the asconcentration parameter. It should be a boolean, but"
                f" {str(type(asconcentration))} was given."
            )
        self.asconcentration = asconcentration

        # effectiveweightformat
        if effectiveweightformat not in [0,1,2,3]: 
            raise ValueError(
                f"{self.__class__.__name__}:" 
                f" Invalid value for effectiveweightformat. This option determines the transformation to"
                f" an equivalent count of particles. Allowed values are 0 (unique effective mass),"
                f" 1 (scale by averaged mass), 2 (real histogram) or 3 (local effective histogram)."
                f" {str(effectiveweightformat)} was given."
            )
        self.effectiveweightformat = effectiveweightformat

        # timepointdata, similar logic to mp7sim
        if timepointdata is not None:
            if not isinstance(timepointdata, (list, tuple)):
                raise TypeError(
                        f"{self.__class__.__name__}:" 
                        f" Invalid type for timepointdata, it"
                        f" should be a list or tuple, but"
                        f" {type(timepoindata)} was given."
                    )
            else:
                if len(timepointdata) != 2:
                    raise ValueError(
                        f"{self.__class__.__name__}:" 
                        f" timepointdata must be a have 2 entries, but "
                        f" {len(timepointdata)} were provided."
                    )
                else:
                    # Infers the specification
                    if isinstance(timepointdata[1], (list, tuple)):
                        timepointdata[1] = np.array(timepointdata[1])
                    elif isinstance(timepointdata[1], (int,float)):
                        timepointdata[1] = np.array([timepointdata[1]]).astype(np.float32)

                    # Infers timepointoption
                    if timepointdata[1].shape[0] == timepointdata[0]:
                        timepointoption = 2
                    elif timepointdata[1].shape[0] > 1:
                        raise ValueError(
                            f"{self.__class__.__name__}:" 
                            f" The number of given timepoints {str(timepointdata[1].shape[0])}"
                            f" is not equal to the given timepointcount {str(timepointdata[0])}"
                        )
                    else:
                        timepointoption = 1
        else:
            # follows timeseries
            timepointoption = 0
        self.timepointoption = timepointoption
        self.timepointdata   = timepointdata

        # skipinitialcondition
        if ( not isinstance( skipinitialcondition, bool ) ): 
            raise TypeError(
                    f"{self.__class__.__name__}:" 
                    f" Invalid type for skipinitialcondition, it"
                    f" should be boolean, but "
                    f" {type(skipinitialcondition)} was given."
                )
        if skipinitialcondition: 
            self.skipinitialcondition = 1
        else:
            self.skipinitialcondition = 0

        # Add package 
        self.parent.add_package(self)


        # And done
        return



    def write_file(self, check=False):
        """
        Write the package file

        Parameters
        ----------
        check : boolean
            Check package data for common errors. (default False)
        Returns
        -------
        None
        """

        # open file for writing
        with open(self.fn_path, "w") as f:

            # line 0:
            # Output file
            if ( ( self.outputcolformat != 0 ) or ( self.outputfileformat != 0 ) ): 
                f.write(f"{self.outputfilename}  {self.outputcolformat}  {self.outputfileformat}\n")
            else:
                f.write(f"{self.outputfilename}\n")

            # line 1:
            # domain origin
            for idb, b in enumerate(self.domainorigin):
                if idb == len(self.domainorigin)-1: 
                    f.write(f"{b:10f}\n")
                else:
                    f.write(f"{b:10f}\t")

            # line 2:
            # domain size
            if ( self.gridallocformat == 1 ):
                if ( self.gridborderfraction > 0 ):
                    for idb, b in enumerate(self.domainsize):
                        f.write(f"{b:10f}\t")
                    f.write(f"{self.gridallocformat}\t")
                    f.write(f"{self.gridborderfraction}\n")
                else:
                    for idb, b in enumerate(self.domainsize):
                        f.write(f"{b:10f}\t")
                    f.write(f"{self.gridallocformat}\n")
            else:
                for idb, b in enumerate(self.domainsize):
                    if idb == len(self.domainsize)-1: 
                        f.write(f"{b:10f}\n")
                    else:
                        f.write(f"{b:10f}\t")

            # line 3:
            # bin sizes
            for idb, b in enumerate(self.binsize):
                if idb == len(self.binsize)-1: 
                    f.write(f"{b:10f}\n")
                else:
                    f.write(f"{b:10f}\t")

            # line 4:
            # sliced reconstruction 
            if not self.slicedreconstruction:
                f.write(f"0\n")
            else:
                f.write(f"1      {self.sliceddimension}\n")

            # line 5:
            # number of optimization loops 
            f.write(f"{self.noptloops}\n")

            # line 6:
            # skip error convergence
            if ( self.skiperror ): 
                f.write(f"{int(self.skiperror)}\n")
            else:
                f.write(f"{int(self.skiperror)}   {self.convergence:6f}\n")

            # line 7:
            # kernels
            f.write(f"{int(self.kerneldatabase)}   {self.boundkernelsize}   {int(self.isotropickernels)}   {int(self.anisotropickernelsic)}\n")

            # line 8:
            # kdb params
            if ( self.kerneldatabase ):
                # Database params: minhd, deltahd, maxhd
                f.write(f"{self.minhd:10f}\t")
                f.write(f"{self.deltahd:10f}\t")
                f.write(f"{self.maxhd:10f}\n")
            elif ( self.boundkernelsize == 1):
                # minhd, maxhd
                f.write(f"{self.minhd:10f}\t")
                f.write(f"{self.maxhd:10f}\n")

            # line 9:
            # initial smoothing
            if ( self.initialsmoothingformat == 1 ):
                f.write(f"{self.initialsmoothingformat}   {self.binsizefactor}\n") 
            else:
                f.write(f"{self.initialsmoothingformat}\n") 

            # line 10
            # as concentration 
            f.write(f"{int(self.asconcentration)}\n")

            # line 11
            # effectiveweightformat
            f.write(f"{self.effectiveweightformat}\n")

            # line 12 
            # timepointoption and skipinitialcondition
            f.write(f"{self.timepointoption}    {self.skipinitialcondition}\n")
           
            # timepointdata
            if self.timepointoption == 1:
                # line 13
                f.write(f"{self.timepointdata[0]}    {self.timepointdata[1][0]}\n")
            elif self.timepointoption == 2:
                # line 13, 14
                f.write(f"{self.timepointdata[0]}\n")
                tp = self.timepointdata[1]
                v  = Util2d(
                    self.parent,
                    (tp.shape[0],),
                    np.float32,
                    tp,
                    name="tpoints",
                    locat=self.unit_number[0],
                )
                f.write(v.string)

        # Done
        return


    def get_output(self):
        '''
        Load the output file into a rec array
        '''
        import os 
        from flopy.utils.flopy_io import loadtxt
        # Shouldn't this be a case to handle by loadtxt ?
        try:
            import pandas as pd
            use_pandas = True
        except ImportError:
            use_pandas = False
       

        if self.outputfileformat == 1:
            raise NotImplementedError(
                f"{self.__class__.__name__}:get_output:" 
                f" Reader for binary output file has not been implemented yet."
            )

        if self.outputcolformat == 0:
            dtype = np.dtype(
                [
                    ("tid"       , np.int32  ),
                    ("time"      , np.float32),
                    ("speciesid" , np.int32  ),
                    ("idbinx"    , np.int32  ),
                    ("idbiny"    , np.int32  ),
                    ("idbinz"    , np.int32  ),
                    ("cgpkde"    , np.float32),
                    ("chist"     , np.float32),
                ]
            )
        elif self.outputcolformat == 1:
            dtype = np.dtype(
                [
                    ("tid"       , np.int32  ),
                    ("time"      , np.float32),
                    ("speciesid" , np.int32  ),
                    ("idbinx"    , np.int32),
                    ("idbiny"    , np.int32),
                    ("idbinz"    , np.int32),
                    ("x"         , np.float32),
                    ("y"         , np.float32),
                    ("z"         , np.float32),
                    ("cgpkde"    , np.float32),
                    ("chist"     , np.float32),
                ]
            )
        elif self.outputcolformat == 2:
            dtype = np.dtype(
                [
                    ("tid"       , np.int32  ),
                    ("time"      , np.float32),
                    ("speciesid" , np.int32  ),
                    ("x"         , np.float32),
                    ("y"         , np.float32),
                    ("z"         , np.float32),
                    ("cgpkde"    , np.float32),
                    ("chist"     , np.float32),
                ]
            )

        # read data
        recdata = loadtxt(
            os.path.join( self._parent.model_ws, self.outputfilename ),
            dtype=dtype,
            skiprows=0, 
            use_pandas=use_pandas,
        )

        # To python zero-based indexes those quantities 
        # that need this treatment. Leave tid as it came.
        recdata['speciesid'] = recdata['speciesid'] - 1
        if self.outputcolformat != 2:
            recdata['idbinx'] = recdata['idbinx'] - 1
            recdata['idbiny'] = recdata['idbiny'] - 1
            recdata['idbinz'] = recdata['idbinz'] - 1

        # store variables
        self.times = np.unique(recdata['time'])
        self.speciesids = np.unique(recdata['speciesid'])

        # useful, heavy ?
        self.outputrecdata = recdata

        # return
        return recdata 


    def get_times(self):
        '''
        Return the array of times stored in data file
        '''
        try:
            return self.times
        except AttributeError:
            self.get_output()
            return self.times


    def get_data(self, which='cgpkde', totim=None, speciesid=None):
        '''
        Get output data array

        For a reconstruction grid coincident with the flowmodel grid of 
        type StructuredGrid and is_regular=True, will fill an array with shape 
        (nlay,nrow,ncol), with the concentration data requested in 'which'.
          * which=cgpkde returns the smoothed reconstruction. 
          * which=chist returns the histogram reconstruction. 

        In case the grid is not of type StructuredGrid or not regular, 
        will filter the recarray data by speciesid and totim, and not by 'which'.
        The same applies for the case in which self.outputcolformat == 2, where 
        reconstruction grid indexes are not given in the output file. 

        Note: the gpkde reconstruction grid follows the modpath convention of 
              coordinates and not the lay,row,col convention so there should an
              adequate reinterpration of grid indexes.
        '''
    
        # Default value for non existent indexes
        defaultnan = 0.0

        # Validate which
        if not isinstance(which,str):
            raise TypeError(
                f"{self.__class__.__name__}:get_data:" 
                f" Invalid type for which parameter, it should be str but {str(type(which))} was given."
            )
        if (which.lower() not in ['cgpkde', 'chist']):
            raise ValueError(
                f"{self.__class__.__name__}:get_data:"
                f" Invalid value for which parameter. It can be"
                f" cgpkde or chist, but {str(which)} was given."
            )
        which = which.lower()

        # Load the output file if not loaded already
        try: 
            recdata = self.outputrecdata
        except AttributeError:
            recdata = self.get_output()

        # If total time was none, assign the last in 
        # self.times
        if totim is None: 
            totim = self.times[-1]
        else:
            tindex = np.where(self.times==totim)[0]
            if len(tindex)==0:
                raise ValueError(
                    f"{self.__class__.__name__}:get_data:"
                    f" The given value for totim was not found in the array of times. "
                    f" totim={str(totim)} was given."
                )
            else:
                totim = self.times[tindex.item()]

        # Similar to times, do it for speciesid
        if speciesid is None: 
            speciesid = self.speciesids[-1]
        else:
            spcindex = np.where(self.speciesids==speciesid)[0]
            if len(spcindex)==0:
                raise ValueError(
                    f"{self.__class__.__name__}:get_data:"
                    f" The given value for speciesid was not found in the array of speciesids. "
                    f" speciesid={str(speciesid)} was given."
                )
            else:
                speciesid = self.speciesids[spcindex.item()]

        # Filter data
        filtdata = recdata[ (recdata['time'] == totim)&(recdata['speciesid'] == speciesid) ]

        # If no data, error.
        if ( len(filtdata) == 0 ):
            raise Exception(
                f"{self.__class__.__name__}:get_data:"
                f" No data was found for totim={str(totim)} and speciesid={str(speciesid)}"
            )

        # If no grid indexes, return
        if ( self.outputcolformat == 2 ):
            # return
            return filtdata

        # The following would only work for
        # StructuredGrid with is_regular=True.
        # Additional alternatives could be provided 
        # by for example interpolating with griddata
        if isinstance( self.parent.flowmodel.modelgrid, StructuredGrid ):

            # If is regular it can potentially be given with 
            # the same structure than the flow model
            if self.parent.flowmodel.modelgrid.is_regular:

                nlay = self.parent.flowmodel.modelgrid.nlay
                nrow = self.parent.flowmodel.modelgrid.nrow
                ncol = self.parent.flowmodel.modelgrid.ncol
                
                data = np.empty((nlay, nrow, ncol), dtype=np.float32)
                data[:, :, :] = defaultnan

                # If the reconstruction grid has the same 
                # size than the flow model grid... 
                # needs checking of consistent sizes. 

                layers = np.unique( filtdata['idbinz'] )
                for lay in layers:
                    srec = filtdata[ filtdata['idbinz'] == lay ]
                    laydata = np.empty((nrow,ncol),dtype=np.float32)
                    laydata[:,:] = defaultnan
                    laydata[ nrow - srec['idbiny'] - 1, srec['idbinx'] ] = srec[which] 
                    data[nlay - lay - 1,:,:] = laydata


                return data

            else:
                print( 
                    f"Warning: data for flowmodel with non-regular StructuredGrid" 
                    f" is returned filtered only by totim and speciesid. The user should apply "
                    f" an adequate coordinates conversion (e.g., scipy.interpolate.griddata)."
                )
                # return
                return filtdata
        else:
            print( 
                f"Warning: data for flowmodel with {str(type(self.parent.flowmodel.modelgrid))} grid "
                f" is returned filtered only by totim and speciesid. The user should apply "
                f" an adequate coordinates conversion. "
            )
            # return
            return filtdata






'''
Configuration of MODPATH-RW 
'''

# python
import os
from pathlib import Path
from subprocess import PIPE, STDOUT, Popen

# flopy
from flopy.modpath import Modpath7
from typing import List, Optional, Tuple, Union
from flopy.mbase import resolve_exe
from flopy.utils import flopy_io


class ModpathRW( Modpath7 ):
    '''
    MODPATH-RW class 

    Extends flopy.modpath.Modpath7 and overloads method write_name_file

    While writing input files with check=True, the existence of the 
    following packages will be verified: 
    
        - ModpathRWSim : the simulation class
        - ModpathRWDsp : dispersion package
        - ModpathRWOpts: random walk options 
        - ModpathRWBas : basic package, porosities
    '''

    def __init__(self, *args, **kwargs):

        # Assign defaults values, prioritizing user provided information
        try:
            kwargs['exe_name']
        except KeyError:
            kwargs['exe_name'] = 'mpathrw'
        try:
            kwargs['simfile_ext']
        except KeyError:
            kwargs['simfile_ext'] = 'mprw' # should be consistent with ModpathRWSim
        try:
            kwargs['modelname']
        except KeyError:
            kwargs['modelname'] = 'mprwsim' 
        try:
            kwargs['flowmodel']
            # If flowmodel was given and not model_ws, 
            # define by default with the model_ws from 
            # the flowmodel
            try:
                kwargs['model_ws']
            except KeyError:
                kwargs['model_ws'] = kwargs['flowmodel'].model_ws
        except KeyError:
            raise ValueError(
                    f"{self.__class__.__name__}:"
                    f" The flowmodel argument should be given for defining a particle tracking model."
                )

        # Call parent constructor
        super().__init__(*args,**kwargs)
        # update version
        self.version_types = {"modpathrw": "MODPATH-RW"}
        self.set_version('modpathrw')


        # Following filenames are generated
        # after parent constructor, appended to the class
        #   - mpbas_file
        #   - dis_file
        #   - grbdis_file
        #   - tdis_file
        #   - headfilename
        #   - budgetfilename

        # Specific MODPATH-RW filenames are 
        # extracted from their respective package ( if given )


    # Overload repr
    def __repr__(self):
        return "MODPATH-RW model"


    # Overload the write_input method
    def write_input(self, SelPackList=False, check=False):
        """
        Write the input.

        Parameters
        ----------
        SelPackList : False or list of packages

        Note: is practically the same than BaseModel but passing
        the check parameter until the write_name_file function
        """
        if check:
            # run check prior to writing input
            self.check(f=f"{self.name}.chk", verbose=self.verbose, level=1)

        # reset the model to free_format if parameter substitution was
        # performed on a model load
        if self.parameter_load and not self.free_format_input:
            if self.verbose:
                print(
                    "\nResetting free_format_input to True to "
                    "preserve the precision of the parameter data."
                )
            self.free_format_input = True

        if self.verbose:
            print("\nWriting packages:")

        if SelPackList == False:
            for p in self.packagelist:
                if self.verbose:
                    print("   Package: ", p.name[0])
                # prevent individual package checks from running after
                # model-level package check above
                # otherwise checks are run twice
                # or the model level check procedure would have to be split up
                # or each package would need a check argument,
                # or default for package level check would have to be False
                try:
                    p.write_file(check=False)
                except TypeError:
                    p.write_file()
        else:
            for pon in SelPackList:
                for i, p in enumerate(self.packagelist):
                    if pon in p.name:
                        if self.verbose:
                            print("   Package: ", p.name[0])
                        try:
                            p.write_file(check=False)
                        except TypeError:
                            p.write_file()
                            break
        if self.verbose:
            print(" ")

        # write name file
        self.write_name_file(check=check)
        # os.chdir(org_dir)



    # Overload the write_name_file method
    def write_name_file(self,check=False):
        """
        Write the name file
        
        check: bool
            Verifies consistency of the model before writing

        Returns
        -------
        None
        """


        with open(os.path.join(self.model_ws, self.mpnamefile), "w") as f:

            # Write the mpnam file
            f.write(f"{self.heading}\n")

            # check consistency 
            if check: 
                # enforce sim
                simpkg   = self.get_package('MPRWSIM')
                isrw     = False
                foundsim = False
                if simpkg is not None:
                    foundsim = True
                    if simpkg.simulationtype >= 5:
                        isrw = True
                if not foundsim: 
                    simpkg = self.get_package('MPSIM')
                    if simpkg is None: 
                        raise Exception( 
                            f"{self.__class__.__name__}:"
                            f" A simulation package is necessary for modpath models,"
                            f" but None was given. Define a ModpathRWSim or Modpath7Sim package."
                        )

                # enforce bas
                baspkg = self.get_package('MPRWBAS')
                if baspkg is None:
                    baspkg = self.get_package('MPBAS')
                    if baspkg is None: 
                        raise Exception( 
                            f"{self.__class__.__name__}:"
                            f" The basic package is necessary for modpath models,"
                            f" but None was given. Define a ModpathRWBas or Modpath7Bas package."
                        )

            if self.mpbas_file is not None:
                f.write(f"MPBAS      {self.mpbas_file}\n")
            if self.dis_file is not None:
                f.write(f"DIS        {self.dis_file}\n")
            if self.grbdis_file is not None:
                f.write(f"{self.grbtag:10s} {self.grbdis_file}\n")
            if self.tdis_file is not None:
                f.write(f"TDIS       {self.tdis_file}\n")
            if self.headfilename is not None:
                f.write(f"HEAD       {self.headfilename}\n")
            if self.budgetfilename is not None:
                f.write(f"BUDGET     {self.budgetfilename}\n")


            # MODPATH-RW specifc files
            # dsp ( enforced if isrw )
            dsppkg = self.get_package('DSP')
            if check:
                if ( (dsppkg is None) and isrw ): 
                    raise Exception( 
                        f"{self.__class__.__name__}:" 
                        f" A dispersion package is required for a random walk model,"
                        f" but None was given. Define a ModpathRWDsp package."
                    )
            if dsppkg is not None:
                f.write(f"DSP        {dsppkg.file_name[0]}\n")
            # rwopts ( enforced if isrw )
            rwoptspkg = self.get_package('RWOPTS')
            if check:
                if ( (rwoptspkg is None) and isrw and check ): 
                    raise Exception( 
                        f"{self.__class__.__name__}:"
                        f" A random walk options package is required for a random walk model," 
                        f" but None was given. Define a ModpathRWOpts package." 
                    )
            if rwoptspkg is not None:
                f.write(f"RWOPTS     {rwoptspkg.file_name[0]}\n")
            # spc
            spcpkg = self.get_package('SPC')
            if spcpkg is not None:
                f.write(f"SPC        {spcpkg.file_name[0]}\n")
            # gpkde
            gpkdepkg = self.get_package('GPKDE')
            if gpkdepkg is not None:
                f.write(f"GPKDE      {gpkdepkg.file_name[0]}\n")
            # ic
            icpkg = self.get_package('IC')
            if icpkg is not None:
                f.write(f"IC         {icpkg.file_name[0]}\n")
            # imp
            imppkg = self.get_package('IMP')
            if imppkg is not None:
                f.write(f"IMP        {imppkg.file_name[0]}\n")
            # src
            srcpkg = self.get_package('SRC')
            if srcpkg is not None:
                f.write(f"SRC        {srcpkg.file_name[0]}\n")
            # obs
            obspkg = self.get_package('OBS')
            if obspkg is not None:
                f.write(f"OBS        {obspkg.file_name[0]}\n")


        # Done
        return


    def run_model(
        self,
        silent=False,
        pause=False,
        report=False,
        normal_msg="normal termination",
        nprocesses=None,
    ) -> Tuple[bool, List[str]]:
        """
        This method will run the model using subprocess.Popen.
        It is based on the original run_model function from BaseModel 
        in flopy. Some additional logic is included in order to handle
        specific use cases from MODPATH-RW.

        Parameters
        ----------
        silent : boolean
            Echo run information to screen (default is True).
        pause : boolean, optional
            Pause upon completion (default is False).
        report : boolean, optional
            Save stdout lines to a list (buff) which is returned
            by the method . (default is False).
        normal_msg : str
            Normal termination message used to determine if the
            run terminated normally. (default is 'normal termination')
        nprocesses : int
            Defines the number of threads to be used in parallel 
            simulations. It should be greater than 0.
    
        Returns
        -------
        success : boolean
        buff : list of lines of stdout
    
        """
   

        # Interpret nprocesses
        cargs = None
        if nprocesses is not None:
            if not isinstance(nprocesses,int):
                raise TypeError(
                    f"{self.__class__.__name__}:"
                    f" Invalid nprocesses parameter of type {str(type(nprocesses))}."
                    f" It should be of type integer."
                )
            if nprocesses <= 0:
                raise ValueError(
                    f"{self.__class__.__name__}:"
                    f" Invalid nprocesses, it should be greater than zero."
                )

            # If valid nprocesses, then format a cargs
            cargs = ["-np", f"{str(nprocesses)}"]
       
        return run_model(
            self.exe_name,
            self.namefile,
            model_ws=self.model_ws,
            silent=silent,
            pause=pause,
            report=report,
            normal_msg=normal_msg,
            cargs=cargs,
        )


def run_model(
    exe_name: Union[str, os.PathLike],
    namefile: Optional[str],
    model_ws: Union[str, os.PathLike] = os.curdir,
    silent=False,
    pause=False,
    report=False,
    processors=None,
    normal_msg="normal termination",
    use_async=False,
    cargs=None,
) -> Tuple[bool, List[str]]:
    """
    This function is mostly the same as the original from flopy, the only 
    difference being that the command arguments (cargs) are passed before the namefile,
    which is the structure in MODPATH-RW.
    
    Original documentation
    ----------------------

    Run the model using subprocess.Popen, optionally collecting stdout and printing
    timestamped progress. Model workspace, namefile, executable to use, and several
    other options may be configured, and additional command line arguments may also
    be provided.

    Parameters
    ----------
    exe_name : str or PathLike
        Executable name or path. If the executable name is provided,
        the executable must be on the system path. Alternatively, a
        full path to the executable may be provided.
    namefile : str, optional
        Name of the name file of model to run. The name may be None
        to run models that don't require a control file (name file)
    model_ws : str or PathLike, optional, default '.'
        Path to the parent directory of the namefile. (default is the
        current working directory '.')
    silent : boolean, default True
        Whether to suppress model output. (Default is True)
    pause : boolean, optional, default False
        Pause and wait for keystroke upon completion. (Default is False)
    report : boolean, optional, default False
        Save stdout lines to a list (buff) returned by the method. (Default is False)
    processors: int
        Number of processors. Parallel simulations are only supported for
        MODFLOW 6 simulations. (default is None)
    normal_msg : str or list
        Termination message used to determine if the model terminated normally.
        More than one message can be provided using a list.
        (Default is 'normal termination')
    use_async : boolean
        Asynchronously read model stdout and report with timestamps. Good for
        models taking a long time to run, not good for models that run quickly.
    cargs : str or list, optional, default None
        Additional command line arguments to pass to the executable.
        (Default is None)
    Returns
    -------
    success : boolean
    buff : list of lines of stdout (empty if report is False)

    """
    success = False
    buff = []

    # convert normal_msg to a list of lower case str for comparison
    if isinstance(normal_msg, str):
        normal_msg = [normal_msg]
    for idx, s in enumerate(normal_msg):
        normal_msg[idx] = s.lower()

    # make sure executable exists
    if exe_name is None:
        raise ValueError(f"An executable name or path must be provided")
    exe_path = resolve_exe(exe_name)
    if not silent:
        print(
            f"FloPy is using the following executable to run the model: {flopy_io.relpath_safe(exe_path, model_ws)}"
        )

    # make sure namefile exists
    if namefile is not None and not os.path.isfile(
        os.path.join(model_ws, namefile)
    ):
        raise FileNotFoundError(
            f"The namefile for this model does not exist: {namefile}"
        )

    # simple little function for the thread to target
    def q_output(output, q):
        for line in iter(output.readline, b""):
            q.put(line)
            # time.sleep(1)
            # output.close()

    # create a list of arguments to pass to Popen
    if processors is not None:
        if "mf6" not in exe_path:
            raise ValueError("processors kwarg only supported for MODFLOW 6")
        mpiexec_path = resolve_exe("mpiexec")
        if not silent:
            print(
                f"FloPy is using {mpiexec_path} "
                + f"to run {exe_path} "
                + f"on {processors} processors."
            )
        argv = [mpiexec_path, "-np", f"{processors}", exe_path, "-p"]
    else:
        argv = [exe_path]

    # This is the difference with the original function in flopy.mbase 

    # add additional arguments to Popen arguments
    if cargs is not None:
        if isinstance(cargs, str):
            cargs = [cargs]
        for t in cargs:
            argv.append(t)

    if namefile is not None:
        argv.append(Path(namefile).name)

    # END This is the difference with the original function in flopy.mbase 

    # run the model with Popen
    proc = Popen(argv, stdout=PIPE, stderr=STDOUT, cwd=model_ws)

    if not use_async:
        while True:
            line = proc.stdout.readline().decode("utf-8")
            if line == "" and proc.poll() is not None:
                break
            if line:
                for msg in normal_msg:
                    if msg in line.lower():
                        success = True
                        break
                line = line.rstrip("\r\n")
                if not silent:
                    print(line)
                if report:
                    buff.append(line)
            else:
                break
        return success, buff

    # some tricks for the async stdout reading
    q = Queue.Queue()
    thread = threading.Thread(target=q_output, args=(proc.stdout, q))
    thread.daemon = True
    thread.start()

    failed_words = ["fail", "error"]
    last = datetime.now()
    lastsec = 0.0
    while True:
        try:
            line = q.get_nowait()
        except Queue.Empty:
            pass
        else:
            if line == "":
                break
            line = line.decode().lower().strip()
            if line != "":
                now = datetime.now()
                dt = now - last
                tsecs = dt.total_seconds() - lastsec
                line = f"(elapsed:{tsecs})-->{line}"
                lastsec = tsecs + lastsec
                buff.append(line)
                if not silent:
                    print(line)
                for fword in failed_words:
                    if fword in line:
                        success = False
                        break
        if proc.poll() is not None:
            break
    proc.wait()
    thread.join(timeout=1)
    buff.extend(proc.stdout.readlines())
    proc.stdout.close()

    for line in buff:
        for msg in normal_msg:
            if msg in line.lower():
                print("success")
                success = True
                break

    if pause:
        input("Press Enter to continue...")
    return success, buff

